@@ -11,6 +11,7 @@ import {
   notify
 } from './common.js';
 import * as Permissions from './permissions.js';
+import * as ApiTabs from '/common/api-tabs.js';
 
 import MenuUI from '/extlib/MenuUI.js';
 import RichConfirm from '/extlib/RichConfirm.js';
@@ -25,7 +26,7 @@ async function getItemById(id) {
   if (!id)
     return null;
   try {
-    const items = await browser.bookmarks.get(id);
+    const items = await browser.bookmarks.get(id).catch(ApiTabs.createErrorHandler());
     if (items.length > 0)
       return items[0];
   }
@@ -42,14 +43,15 @@ export async function bookmarkTab(tab, options = {}) {
   catch(_e) {
     notify({
       title:   browser.i18n.getMessage('bookmark_notification_notPermitted_title'),
-      message: browser.i18n.getMessage('bookmark_notification_notPermitted_message')
+      message: browser.i18n.getMessage('bookmark_notification_notPermitted_message'),
+      url:     `moz-extension://${location.host}/options/options.html#bookmarksPermissionSection`
     });
     return null;
   }
   const parent = await getItemById(options.parentId || configs.defaultBookmarkParentId);
 
-  let title    = tab.apiTab.title;
-  let url      = tab.apiTab.url;
+  let title    = tab.title;
+  let url      = tab.url;
   let parentId = parent && parent.id;
   if (options.showDialog) {
     try {
@@ -72,6 +74,7 @@ export async function bookmarkTab(tab, options = {}) {
           initFolderChoolser(container.querySelector('button'), {
             defaultValue: parentId
           });
+          container.querySelector('[name="title"]').select();
         },
         buttons: [
           browser.i18n.getMessage('bookmarkDialog_accept'),
@@ -91,7 +94,7 @@ export async function bookmarkTab(tab, options = {}) {
 
   const item = await browser.bookmarks.create({
     parentId, title, url
-  });
+  }).catch(ApiTabs.createErrorHandler());
   return item;
 }
 
@@ -103,14 +106,15 @@ export async function bookmarkTabs(tabs, options = {}) {
   catch(_e) {
     notify({
       title:   browser.i18n.getMessage('bookmark_notification_notPermitted_title'),
-      message: browser.i18n.getMessage('bookmark_notification_notPermitted_message')
+      message: browser.i18n.getMessage('bookmark_notification_notPermitted_message'),
+      url:     `moz-extension://${location.host}/options/options.html#bookmarksPermissionSection`
     });
     return null;
   }
   const now = new Date();
   const title = configs.bookmarkTreeFolderName
-    .replace(/%TITLE%/gi, tabs[0].apiTab.title)
-    .replace(/%URL%/gi, tabs[0].apiTab.url)
+    .replace(/%TITLE%/gi, tabs[0].title)
+    .replace(/%URL%/gi, tabs[0].url)
     .replace(/%YEAR%/gi, now.getFullYear())
     .replace(/%MONTH%/gi, `0${now.getMonth() + 1}`.substr(-2))
     .replace(/%DATE%/gi, `0${now.getDate()}`.substr(-2));
@@ -144,6 +148,7 @@ export async function bookmarkTabs(tabs, options = {}) {
           initFolderChoolser(container.querySelector('button'), {
             defaultValue: folderParams.parentId
           });
+          container.querySelector('[name="title"]').select();
         },
         buttons: [
           browser.i18n.getMessage('bookmarkDialog_accept'),
@@ -160,15 +165,15 @@ export async function bookmarkTabs(tabs, options = {}) {
     }
   }
 
-  const folder = await browser.bookmarks.create(folderParams);
+  const folder = await browser.bookmarks.create(folderParams).catch(ApiTabs.createErrorHandler());
   for (let i = 0, maxi = tabs.length; i < maxi; i++) {
     const tab = tabs[i];
     await browser.bookmarks.create({
       parentId: folder.id,
       index:    i,
-      title:    tab.apiTab.title,
-      url:      tab.apiTab.url
-    });
+      title:    tab.title,
+      url:      tab.url
+    }).catch(ApiTabs.createErrorSuppressor());
   }
   return folder;
 }
@@ -276,6 +281,6 @@ export async function initFolderChoolser(anchor, params = {}) {
     }
   };
 
-  const rootItems = await browser.bookmarks.getTree();
+  const rootItems = await browser.bookmarks.getTree().catch(ApiTabs.createErrorHandler());
   buildItems(rootItems[0].children, mChooserTree);
 }
