@@ -11,12 +11,12 @@ import {
 } from '/common/common.js';
 
 import * as Constants from '/common/constants.js';
+import * as ApiTabs from '/common/api-tabs.js';
 
 function log(...args) {
   internalLogger('background/browser-action-menu', ...args);
 }
 
-const isMac = /^Mac/i.test(navigator.platform);
 const delimiter = browser.i18n.getMessage('config_terms_delimiter');
 
 function indent(level = 1) {
@@ -119,54 +119,18 @@ const mItems = [
         ]
       },
       {
-        title:    browser.i18n.getMessage('config_scrollbarMode_caption'),
+        title:    browser.i18n.getMessage('config_labelOverflowStyle_caption'),
         children: [
           {
-            title: browser.i18n.getMessage('config_scrollbarMode_default'),
-            key:   'scrollbarMode',
-            value: Constants.kTABBAR_SCROLLBAR_MODE_DEFAULT,
+            title: browser.i18n.getMessage('config_labelOverflowStyle_fade'),
+            key:   'labelOverflowStyle',
+            value: 'fade',
             type:  'radio'
           },
           {
-            title: browser.i18n.getMessage('config_scrollbarMode_narrow'),
-            key:   'scrollbarMode',
-            value: Constants.kTABBAR_SCROLLBAR_MODE_NARROW,
-            type:  'radio'
-          },
-          {
-            title: browser.i18n.getMessage('config_scrollbarMode_overlay'),
-            key:   'scrollbarMode',
-            value: Constants.kTABBAR_SCROLLBAR_MODE_OVERLAY,
-            type:  'radio',
-            visible: isMac
-          },
-          {
-            title: browser.i18n.getMessage('config_scrollbarMode_hide'),
-            key:   'scrollbarMode',
-            value: Constants.kTABBAR_SCROLLBAR_MODE_HIDE,
-            type:  'radio'
-          }
-        ]
-      },
-      {
-        title:    browser.i18n.getMessage('config_sidebarScrollbarPosition_caption'),
-        children: [
-          {
-            title: browser.i18n.getMessage('config_sidebarScrollbarPosition_auto'),
-            key:   'sidebarScrollbarPosition',
-            value: Constants.kTABBAR_SCROLLBAR_POSITION_AUTO,
-            type:  'radio'
-          },
-          {
-            title: browser.i18n.getMessage('config_sidebarScrollbarPosition_left'),
-            key:   'sidebarScrollbarPosition',
-            value: Constants.kTABBAR_SCROLLBAR_POSITION_LEFT,
-            type:  'radio'
-          },
-          {
-            title: browser.i18n.getMessage('config_sidebarScrollbarPosition_right'),
-            key:   'sidebarScrollbarPosition',
-            value: Constants.kTABBAR_SCROLLBAR_POSITION_RIGHT,
+            title: browser.i18n.getMessage('config_labelOverflowStyle_crop'),
+            key:   'labelOverflowStyle',
+            value: 'crop',
             type:  'radio'
           }
         ]
@@ -197,75 +161,122 @@ const mItems = [
     title:    browser.i18n.getMessage('config_context_caption'),
     children: [
       {
-        title: browser.i18n.getMessage('context_topLevel_prefix') + browser.i18n.getMessage('context_closeTree_label'),
-        key:   'context_closeTabOptions_closeTree',
+        title: browser.i18n.getMessage('config_extraItems_topLevel'),
+        enabled: false
+      },
+      {
+        title: indent() + browser.i18n.getMessage('context_reloadTree_label'),
+        key:   'context_topLevel_reloadTree',
         type:  'checkbox'
       },
       {
-        title: browser.i18n.getMessage('context_topLevel_prefix') + browser.i18n.getMessage('context_closeDescendants_label'),
-        key:   'context_closeTabOptions_closeDescendants',
+        title: indent() + browser.i18n.getMessage('context_reloadDescendants_label'),
+        key:   'context_topLevel_reloadDescendants',
         type:  'checkbox'
       },
       {
-        title: browser.i18n.getMessage('context_topLevel_prefix') + browser.i18n.getMessage('context_closeOthers_label'),
-        key:   'context_closeTabOptions_closeOthers',
+        title: indent() + browser.i18n.getMessage('context_closeTree_label'),
+        key:   'context_topLevel_closeTree',
+        type:  'checkbox'
+      },
+      {
+        title: indent() + browser.i18n.getMessage('context_closeDescendants_label'),
+        key:   'context_topLevel_closeDescendants',
+        type:  'checkbox'
+      },
+      {
+        title: indent() + browser.i18n.getMessage('context_closeOthers_label'),
+        key:   'context_topLevel_closeOthers',
+        type:  'checkbox'
+      },
+      {
+        title: indent() + browser.i18n.getMessage('context_collapseTree_label'),
+        key:   'context_topLevel_collapseTree',
+        type:  'checkbox'
+      },
+      {
+        title: indent() + browser.i18n.getMessage('context_collapseAll_label'),
+        key:   'context_topLevel_collapseAll',
+        type:  'checkbox'
+      },
+      {
+        title: indent() + browser.i18n.getMessage('context_expandTree_label'),
+        key:   'context_topLevel_expandTree',
+        type:  'checkbox'
+      },
+      {
+        title: indent() + browser.i18n.getMessage('context_expandAll_label'),
+        key:   'context_topLevel_expandAll',
+        type:  'checkbox'
+      },
+      {
+        title: indent() + browser.i18n.getMessage('context_bookmarkTree_label'),
+        key:   'context_topLevel_bookmarkTree',
+        type:  'checkbox'
+      },
+      {
+        title: indent() + browser.i18n.getMessage('context_groupTabs_label'),
+        key:   'context_topLevel_groupTabs',
         type:  'checkbox'
       },
       { type: 'separator' },
       {
-        title: browser.i18n.getMessage('context_reloadTree_label'),
+        title: browser.i18n.getMessage('config_extraItems_subMenu'),
+        enabled: false
+      },
+      {
+        title: indent() + browser.i18n.getMessage('context_reloadTree_label'),
         key:   'context_reloadTree',
         type:  'checkbox'
       },
       {
-        title: browser.i18n.getMessage('context_reloadDescendants_label'),
+        title: indent() + browser.i18n.getMessage('context_reloadDescendants_label'),
         key:   'context_reloadDescendants',
         type:  'checkbox'
       },
       {
-        title: browser.i18n.getMessage('context_closeTree_label'),
+        title: indent() + browser.i18n.getMessage('context_closeTree_label'),
         key:   'context_closeTree',
         type:  'checkbox'
       },
       {
-        title: browser.i18n.getMessage('context_closeDescendants_label'),
+        title: indent() + browser.i18n.getMessage('context_closeDescendants_label'),
         key:   'context_closeDescendants',
         type:  'checkbox'
       },
       {
-        title: browser.i18n.getMessage('context_closeOthers_label'),
+        title: indent() + browser.i18n.getMessage('context_closeOthers_label'),
         key:   'context_closeOthers',
         type:  'checkbox'
       },
       {
-        title: browser.i18n.getMessage('context_collapseTree_label'),
+        title: indent() + browser.i18n.getMessage('context_collapseTree_label'),
         key:   'context_collapseTree',
         type:  'checkbox'
       },
       {
-        title: browser.i18n.getMessage('context_collapseAll_label'),
+        title: indent() + browser.i18n.getMessage('context_collapseAll_label'),
         key:   'context_collapseAll',
         type:  'checkbox'
       },
       {
-        title: browser.i18n.getMessage('context_expandTree_label'),
+        title: indent() + browser.i18n.getMessage('context_expandTree_label'),
         key:   'context_expandTree',
         type:  'checkbox'
       },
       {
-        title: browser.i18n.getMessage('context_expandAll_label'),
+        title: indent() + browser.i18n.getMessage('context_expandAll_label'),
         key:   'context_expandAll',
         type:  'checkbox'
       },
       {
-        title: browser.i18n.getMessage('context_bookmarkTree_label'),
+        title: indent() + browser.i18n.getMessage('context_bookmarkTree_label'),
         key:   'context_bookmarkTree',
         type:  'checkbox'
       },
-      { type: 'separator' },
       {
-        title: browser.i18n.getMessage('config_fakeContextMenu_label'),
-        key:   'fakeContextMenu',
+        title: indent() + browser.i18n.getMessage('context_groupTabs_label'),
+        key:   'context_groupTabs',
         type:  'checkbox'
       }
     ]
@@ -279,6 +290,12 @@ const mItems = [
       {
         title: indent() + browser.i18n.getMessage('config_autoAttachOnNewTabButtonMiddleClick_before'),
         children: [
+          {
+            title: browser.i18n.getMessage('config_autoAttachOnNewTabButtonMiddleClick_noControl') + delimiter + browser.i18n.getMessage('config_autoAttachOnNewTabButtonMiddleClick_after'),
+            key:   'autoAttachOnNewTabButtonMiddleClick',
+            value: Constants.kNEWTAB_DO_NOTHING,
+            type:  'radio'
+          },
           {
             title: browser.i18n.getMessage('config_autoAttachOnNewTabButtonMiddleClick_independent') + delimiter + browser.i18n.getMessage('config_autoAttachOnNewTabButtonMiddleClick_after'),
             key:   'autoAttachOnNewTabButtonMiddleClick',
@@ -301,8 +318,7 @@ const mItems = [
             title: browser.i18n.getMessage('config_autoAttachOnNewTabButtonMiddleClick_nextSibling') + delimiter + browser.i18n.getMessage('config_autoAttachOnNewTabButtonMiddleClick_after'),
             key:   'autoAttachOnNewTabButtonMiddleClick',
             value: Constants.kNEWTAB_OPEN_AS_NEXT_SIBLING,
-            type:  'radio',
-            visible: isMac
+            type:  'radio'
           }
         ]
       },
@@ -341,11 +357,18 @@ const mItems = [
       },
       { type: 'separator' },
       {
-        title: browser.i18n.getMessage('config_newTabAction_caption')
+        title: browser.i18n.getMessage('config_newTabAction_caption'),
+        enabled: false
       },
       {
         title:    indent() + browser.i18n.getMessage('config_autoAttachOnNewTabCommand_before'),
         children: [
+          {
+            title: browser.i18n.getMessage('config_autoAttachOnNewTabCommand_noControl') + delimiter + browser.i18n.getMessage('config_autoAttachOnNewTabCommand_after'),
+            key:   'autoAttachOnNewTabCommand',
+            value: Constants.kNEWTAB_DO_NOTHING,
+            type:  'radio'
+          },
           {
             title: browser.i18n.getMessage('config_autoAttachOnNewTabCommand_independent') + delimiter + browser.i18n.getMessage('config_autoAttachOnNewTabCommand_after'),
             key:   'autoAttachOnNewTabCommand',
@@ -368,14 +391,19 @@ const mItems = [
             title: browser.i18n.getMessage('config_autoAttachOnNewTabCommand_nextSibling') + delimiter + browser.i18n.getMessage('config_autoAttachOnNewTabCommand_after'),
             key:   'autoAttachOnNewTabCommand',
             value: Constants.kNEWTAB_OPEN_AS_NEXT_SIBLING,
-            type:  'radio',
-            visible: isMac
+            type:  'radio'
           }
         ]
       },
       {
         title:    indent() + browser.i18n.getMessage('config_autoAttachOnDuplicated_before'),
         children: [
+          {
+            title: browser.i18n.getMessage('config_autoAttachOnDuplicated_noControl') + delimiter + browser.i18n.getMessage('config_autoAttachOnDuplicated_after'),
+            key:   'autoAttachOnDuplicated',
+            value: Constants.kNEWTAB_DO_NOTHING,
+            type:  'radio'
+          },
           {
             title: browser.i18n.getMessage('config_autoAttachOnDuplicated_independent') + delimiter + browser.i18n.getMessage('config_autoAttachOnDuplicated_after'),
             key:   'autoAttachOnDuplicated',
@@ -398,8 +426,7 @@ const mItems = [
             title: browser.i18n.getMessage('config_autoAttachOnDuplicated_nextSibling') + delimiter + browser.i18n.getMessage('config_autoAttachOnDuplicated_after'),
             key:   'autoAttachOnDuplicated',
             value: Constants.kNEWTAB_OPEN_AS_NEXT_SIBLING,
-            type:  'radio',
-            visible: isMac
+            type:  'radio'
           }
         ]
       },
@@ -412,6 +439,12 @@ const mItems = [
       {
         title:    browser.i18n.getMessage('config_sameSiteOrphan_caption'),
         children: [
+          {
+            title: browser.i18n.getMessage('config_autoAttachSameSiteOrphan_before') + delimiter + browser.i18n.getMessage('config_autoAttachSameSiteOrphan_noControl') + delimiter + browser.i18n.getMessage('config_autoAttachSameSiteOrphan_after'),
+            key:   'autoAttachSameSiteOrphan',
+            value: Constants.kNEWTAB_DO_NOTHING,
+            type:  'radio'
+          },
           {
             title: browser.i18n.getMessage('config_autoAttachSameSiteOrphan_before') + delimiter + browser.i18n.getMessage('config_autoAttachSameSiteOrphan_independent') + delimiter + browser.i18n.getMessage('config_autoAttachSameSiteOrphan_after'),
             key:   'autoAttachSameSiteOrphan',
@@ -434,8 +467,7 @@ const mItems = [
             title: browser.i18n.getMessage('config_autoAttachSameSiteOrphan_before') + delimiter + browser.i18n.getMessage('config_autoAttachSameSiteOrphan_nextSibling') + delimiter + browser.i18n.getMessage('config_autoAttachSameSiteOrphan_after'),
             key:   'autoAttachSameSiteOrphan',
             value: Constants.kNEWTAB_OPEN_AS_NEXT_SIBLING,
-            type:  'radio',
-            visible: isMac
+            type:  'radio'
           },
           { type: 'separator' },
           {
@@ -471,12 +503,6 @@ const mItems = [
       {
         title:    browser.i18n.getMessage('config_insertNewTabFromPinnedTabAt_caption'),
         children: [
-          {
-            title: browser.i18n.getMessage('config_insertNewTabFromPinnedTabAt_caption'),
-            key:   'autoGroupNewTabsFromPinned',
-            type:  'checkbox'
-          },
-          { type:  'separator' },
           {
             title: browser.i18n.getMessage('config_insertNewTabFromPinnedTabAt_noControl'),
             key:   'insertNewTabFromPinnedTabAt',
@@ -497,13 +523,28 @@ const mItems = [
           }
         ]
       },
+      { type: 'separator' },
+      {
+        title: browser.i18n.getMessage('config_groupTab_caption'),
+        enabled: false
+      },
       {
         dynamicTitle: true,
         get title() {
-          return browser.i18n.getMessage('config_autoGroupNewTabs_before') + delimiter + configs.autoGroupNewTabsTimeout + delimiter + browser.i18n.getMessage('config_autoGroupNewTabs_after');
+          return indent() + browser.i18n.getMessage('config_autoGroupNewTabs_before') + delimiter + configs.autoGroupNewTabsTimeout + delimiter + browser.i18n.getMessage('config_autoGroupNewTabs_after');
         },
         key:   'autoGroupNewTabs',
         type:  'checkbox'
+      },
+      {
+        title: indent() + browser.i18n.getMessage('config_insertNewTabFromPinnedTabAt_caption'),
+        key:   'autoGroupNewTabsFromPinned',
+        type:  'checkbox'
+      },
+      {
+        title: indent() + browser.i18n.getMessage('config_renderTreeInGroupTabs_label'),
+        key:   'renderTreeInGroupTabs',
+        type:  'checkbox'
       }
     ]
   },
@@ -562,64 +603,227 @@ const mItems = [
         }
       },
       {
-        title:    browser.i18n.getMessage('config_closeParentBehavior_caption'),
+        title:    browser.i18n.getMessage('config_closeParentBehaviorMode_caption'),
         children: [
           {
-            title: browser.i18n.getMessage('config_closeParentBehavior_close'),
-            key:   'closeParentBehavior',
-            value: Constants.kCLOSE_PARENT_BEHAVIOR_CLOSE_ALL_CHILDREN,
+            title: browser.i18n.getMessage('config_closeParentBehaviorMode_withNativeTabs'),
+            key:   'closeParentBehaviorMode',
+            value: Constants.kCLOSE_PARENT_BEHAVIOR_MODE_WITH_NATIVE_TABBAR,
             type:  'radio'
           },
           {
-            title: browser.i18n.getMessage('config_closeParentBehavior_replaceWithGroupTab'),
-            key:   'closeParentBehavior',
-            value: Constants.kCLOSE_PARENT_BEHAVIOR_REPLACE_WITH_GROUP_TAB,
+            title: indent() + browser.i18n.getMessage('config_closeParentBehaviorMode_withNativeTabs_caption_withoutSidebar')
+          },
+          {
+            title: indent() + browser.i18n.getMessage('config_closeParentBehaviorMode_withNativeTabs_caption_withSidebar'),
+            children: [
+              {
+                title: browser.i18n.getMessage('config_closeParentBehavior_replaceWithGroupTab'),
+                key:   'closeParentBehavior',
+                value: Constants.kCLOSE_PARENT_BEHAVIOR_REPLACE_WITH_GROUP_TAB,
+                type:  'radio'
+              },
+              {
+                title: browser.i18n.getMessage('config_closeParentBehavior_promoteFirst'),
+                key:   'closeParentBehavior',
+                value: Constants.kCLOSE_PARENT_BEHAVIOR_PROMOTE_FIRST_CHILD,
+                type:  'radio'
+              },
+              {
+                title: browser.i18n.getMessage('config_closeParentBehavior_promoteAll'),
+                key:   'closeParentBehavior',
+                value: Constants.kCLOSE_PARENT_BEHAVIOR_PROMOTE_ALL_CHILDREN,
+                type:  'radio'
+              },
+              {
+                title: browser.i18n.getMessage('config_closeParentBehavior_promoteIntelligently'),
+                key:   'closeParentBehavior',
+                value: Constants.kCLOSE_PARENT_BEHAVIOR_PROMOTE_INTELLIGENTLY,
+                type:  'radio'
+              }
+            ]
+          },
+          { type: 'separator' },
+          {
+            title: browser.i18n.getMessage('config_closeParentBehaviorMode_withoutNativeTabs'),
+            key:   'closeParentBehaviorMode',
+            value: Constants.kCLOSE_PARENT_BEHAVIOR_MODE_WITHOUT_NATIVE_TABBAR,
             type:  'radio'
           },
           {
-            title: browser.i18n.getMessage('config_closeParentBehavior_promoteFirst'),
-            key:   'closeParentBehavior',
-            value: Constants.kCLOSE_PARENT_BEHAVIOR_PROMOTE_FIRST_CHILD,
+            title: indent() + browser.i18n.getMessage('config_closeParentBehaviorMode_withoutNativeTabs_caption'),
+            children: [
+              {
+                title: browser.i18n.getMessage('config_closeParentBehavior_replaceWithGroupTab'),
+                key:   'closeParentBehavior',
+                value: Constants.kCLOSE_PARENT_BEHAVIOR_REPLACE_WITH_GROUP_TAB,
+                type:  'radio'
+              },
+              {
+                title: browser.i18n.getMessage('config_closeParentBehavior_promoteFirst'),
+                key:   'closeParentBehavior',
+                value: Constants.kCLOSE_PARENT_BEHAVIOR_PROMOTE_FIRST_CHILD,
+                type:  'radio'
+              },
+              {
+                title: browser.i18n.getMessage('config_closeParentBehavior_promoteAll'),
+                key:   'closeParentBehavior',
+                value: Constants.kCLOSE_PARENT_BEHAVIOR_PROMOTE_ALL_CHILDREN,
+                type:  'radio'
+              },
+              {
+                title: browser.i18n.getMessage('config_closeParentBehavior_promoteIntelligently'),
+                key:   'closeParentBehavior',
+                value: Constants.kCLOSE_PARENT_BEHAVIOR_PROMOTE_INTELLIGENTLY,
+                type:  'radio'
+              }
+            ]
+          },
+          { type: 'separator' },
+          {
+            title: browser.i18n.getMessage('config_closeParentBehaviorMode_custom'),
+            key:   'closeParentBehaviorMode',
+            value: Constants.kCLOSE_PARENT_BEHAVIOR_MODE_CUSTOM,
             type:  'radio'
           },
           {
-            title: browser.i18n.getMessage('config_closeParentBehavior_promoteAll'),
-            key:   'closeParentBehavior',
-            value: Constants.kCLOSE_PARENT_BEHAVIOR_PROMOTE_ALL_CHILDREN,
-            type:  'radio'
+            title: indent() + browser.i18n.getMessage('config_closeParentBehaviorMode_custom_caption_insideSidebar'),
+            children: [
+              {
+                title: browser.i18n.getMessage('config_closeParentBehavior_close'),
+                key:   'closeParentBehavior',
+                value: Constants.kCLOSE_PARENT_BEHAVIOR_CLOSE_ALL_CHILDREN,
+                type:  'radio'
+              },
+              {
+                title: browser.i18n.getMessage('config_closeParentBehavior_replaceWithGroupTab'),
+                key:   'closeParentBehavior',
+                value: Constants.kCLOSE_PARENT_BEHAVIOR_REPLACE_WITH_GROUP_TAB,
+                type:  'radio'
+              },
+              {
+                title: browser.i18n.getMessage('config_closeParentBehavior_promoteFirst'),
+                key:   'closeParentBehavior',
+                value: Constants.kCLOSE_PARENT_BEHAVIOR_PROMOTE_FIRST_CHILD,
+                type:  'radio'
+              },
+              {
+                title: browser.i18n.getMessage('config_closeParentBehavior_promoteAll'),
+                key:   'closeParentBehavior',
+                value: Constants.kCLOSE_PARENT_BEHAVIOR_PROMOTE_ALL_CHILDREN,
+                type:  'radio'
+              },
+              {
+                title: browser.i18n.getMessage('config_closeParentBehavior_promoteIntelligently'),
+                key:   'closeParentBehavior',
+                value: Constants.kCLOSE_PARENT_BEHAVIOR_PROMOTE_INTELLIGENTLY,
+                type:  'radio'
+              },
+              {
+                title: browser.i18n.getMessage('config_closeParentBehavior_detach'),
+                key:   'closeParentBehavior',
+                value: Constants.kCLOSE_PARENT_BEHAVIOR_DETACH_ALL_CHILDREN,
+                type:  'radio'
+              }
+            ]
           },
           {
-            title: indent() + browser.i18n.getMessage('config_promoteFirstChildForClosedRoot_label'),
-            key:   'promoteFirstChildForClosedRoot',
-            type:  'checkbox'
+            title: indent() + browser.i18n.getMessage('config_closeParentBehaviorMode_custom_caption_outsideSidebar'),
+            children: [
+              {
+                title: browser.i18n.getMessage('config_closeParentBehavior_close'),
+                key:   'closeParentBehavior_outsideSidebar',
+                value: Constants.kCLOSE_PARENT_BEHAVIOR_CLOSE_ALL_CHILDREN,
+                type:  'radio'
+              },
+              {
+                title: browser.i18n.getMessage('config_closeParentBehavior_replaceWithGroupTab'),
+                key:   'closeParentBehavior_outsideSidebar',
+                value: Constants.kCLOSE_PARENT_BEHAVIOR_REPLACE_WITH_GROUP_TAB,
+                type:  'radio'
+              },
+              {
+                title: browser.i18n.getMessage('config_closeParentBehavior_promoteFirst'),
+                key:   'closeParentBehavior_outsideSidebar',
+                value: Constants.kCLOSE_PARENT_BEHAVIOR_PROMOTE_FIRST_CHILD,
+                type:  'radio'
+              },
+              {
+                title: browser.i18n.getMessage('config_closeParentBehavior_promoteAll'),
+                key:   'closeParentBehavior_outsideSidebar',
+                value: Constants.kCLOSE_PARENT_BEHAVIOR_PROMOTE_ALL_CHILDREN,
+                type:  'radio'
+              },
+              {
+                title: browser.i18n.getMessage('config_closeParentBehavior_promoteIntelligently'),
+                key:   'closeParentBehavior_outsideSidebar',
+                value: Constants.kCLOSE_PARENT_BEHAVIOR_PROMOTE_INTELLIGENTLY,
+                type:  'radio'
+              },
+              {
+                title: browser.i18n.getMessage('config_closeParentBehavior_detach'),
+                key:   'closeParentBehavior_outsideSidebar',
+                value: Constants.kCLOSE_PARENT_BEHAVIOR_DETACH_ALL_CHILDREN,
+                type:  'radio'
+              }
+            ]
           },
           {
-            title: browser.i18n.getMessage('config_closeParentBehavior_detach'),
-            key:   'closeParentBehavior',
-            value: Constants.kCLOSE_PARENT_BEHAVIOR_DETACH_ALL_CHILDREN,
-            type:  'radio'
+            title: indent() + browser.i18n.getMessage('config_closeParentBehaviorMode_custom_caption_noSidebar'),
+            children: [
+              {
+                title: browser.i18n.getMessage('config_closeParentBehavior_close'),
+                key:   'closeParentBehavior_noSidebar',
+                value: Constants.kCLOSE_PARENT_BEHAVIOR_CLOSE_ALL_CHILDREN,
+                type:  'radio'
+              },
+              {
+                title: browser.i18n.getMessage('config_closeParentBehavior_replaceWithGroupTab'),
+                key:   'closeParentBehavior_noSidebar',
+                value: Constants.kCLOSE_PARENT_BEHAVIOR_REPLACE_WITH_GROUP_TAB,
+                type:  'radio'
+              },
+              {
+                title: browser.i18n.getMessage('config_closeParentBehavior_promoteFirst'),
+                key:   'closeParentBehavior_noSidebar',
+                value: Constants.kCLOSE_PARENT_BEHAVIOR_PROMOTE_FIRST_CHILD,
+                type:  'radio'
+              },
+              {
+                title: browser.i18n.getMessage('config_closeParentBehavior_promoteAll'),
+                key:   'closeParentBehavior_noSidebar',
+                value: Constants.kCLOSE_PARENT_BEHAVIOR_PROMOTE_ALL_CHILDREN,
+                type:  'radio'
+              },
+              {
+                title: browser.i18n.getMessage('config_closeParentBehavior_promoteIntelligently'),
+                key:   'closeParentBehavior_noSidebar',
+                value: Constants.kCLOSE_PARENT_BEHAVIOR_PROMOTE_INTELLIGENTLY,
+                type:  'radio'
+              },
+              {
+                title: browser.i18n.getMessage('config_closeParentBehavior_detach'),
+                key:   'closeParentBehavior_noSidebar',
+                value: Constants.kCLOSE_PARENT_BEHAVIOR_DETACH_ALL_CHILDREN,
+                type:  'radio'
+              }
+            ]
           }
         ]
       },
       {
-        title:    browser.i18n.getMessage('config_parentTabBehaviorForChanges_caption'),
+        title:    browser.i18n.getMessage('config_fixupTreeOnTabVisibilityChanged_caption'),
         children: [
           {
-            title: browser.i18n.getMessage('config_parentTabBehaviorForChanges_always'),
-            key:   'parentTabBehaviorForChanges',
-            value: Constants.kPARENT_TAB_BEHAVIOR_ALWAYS,
+            title: browser.i18n.getMessage('config_fixupTreeOnTabVisibilityChanged_fix'),
+            key:   'fixupTreeOnTabVisibilityChanged',
+            value: true,
             type:  'radio'
           },
           {
-            title: browser.i18n.getMessage('config_parentTabBehaviorForChanges_alwaysButOnlyWhenVisible'),
-            key:   'parentTabBehaviorForChanges',
-            value: Constants.kPARENT_TAB_BEHAVIOR_ONLY_WHEN_VISIBLE,
-            type:  'radio'
-          },
-          {
-            title: browser.i18n.getMessage('config_parentTabBehaviorForChanges_onlyInSidebar'),
-            key:   'parentTabBehaviorForChanges',
-            value: Constants.kPARENT_TAB_BEHAVIOR_ONLY_ON_SIDEBAR,
+            title: browser.i18n.getMessage('config_fixupTreeOnTabVisibilityChanged_keep'),
+            key:   'fixupTreeOnTabVisibilityChanged',
+            value: false,
             type:  'radio'
           }
         ]
@@ -630,7 +834,8 @@ const mItems = [
     title:    browser.i18n.getMessage('config_drag_caption'),
     children: [
       {
-        title: browser.i18n.getMessage('config_tabDragBehavior_caption')
+        title: browser.i18n.getMessage('config_tabDragBehavior_caption'),
+        enabled: false
       },
       {
         title:    indent() + browser.i18n.getMessage('config_tabDragBehavior_label'),
@@ -638,7 +843,7 @@ const mItems = [
           {
             title: browser.i18n.getMessage('config_tabDragBehavior_tearoff_tree'),
             key:   'tabDragBehavior',
-            value: Constants.kDRAG_BEHAVIOR_WHOLE_TREE,
+            value: Constants.kDRAG_BEHAVIOR_TEAR_OFF | Constants.kDRAG_BEHAVIOR_WHOLE_TREE,
             type:  'radio'
           },
           {
@@ -650,7 +855,7 @@ const mItems = [
           {
             title: browser.i18n.getMessage('config_tabDragBehavior_tearoff_tab'),
             key:   'tabDragBehavior',
-            value: Constants.kDRAG_BEHAVIOR_NONE,
+            value: Constants.kDRAG_BEHAVIOR_TEAR_OFF,
             type:  'radio'
           },
           {
@@ -658,6 +863,12 @@ const mItems = [
             key:   'tabDragBehavior',
             value: Constants.kDRAG_BEHAVIOR_ALLOW_BOOKMARK,
             type:  'radio'
+          },
+          {
+            title: browser.i18n.getMessage('config_tabDragBehavior_none'),
+            key:   'tabDragBehavior',
+            value: Constants.kDRAG_BEHAVIOR_NONE,
+            type:  'radio'
           }
         ]
       },
@@ -667,7 +878,7 @@ const mItems = [
           {
             title: browser.i18n.getMessage('config_tabDragBehavior_tearoff_tree'),
             key:   'tabDragBehaviorShift',
-            value: Constants.kDRAG_BEHAVIOR_WHOLE_TREE,
+            value: Constants.kDRAG_BEHAVIOR_TEAR_OFF | Constants.kDRAG_BEHAVIOR_WHOLE_TREE,
             type:  'radio'
           },
           {
@@ -679,7 +890,7 @@ const mItems = [
           {
             title: browser.i18n.getMessage('config_tabDragBehavior_tearoff_tab'),
             key:   'tabDragBehaviorShift',
-            value: Constants.kDRAG_BEHAVIOR_NONE,
+            value: Constants.kDRAG_BEHAVIOR_TEAR_OFF,
             type:  'radio'
           },
           {
@@ -687,6 +898,12 @@ const mItems = [
             key:   'tabDragBehaviorShift',
             value: Constants.kDRAG_BEHAVIOR_ALLOW_BOOKMARK,
             type:  'radio'
+          },
+          {
+            title: browser.i18n.getMessage('config_tabDragBehavior_none'),
+            key:   'tabDragBehaviorShift',
+            value: Constants.kDRAG_BEHAVIOR_NONE,
+            type:  'radio'
           }
         ]
       },
@@ -749,6 +966,11 @@ const mItems = [
         key:   'supportTabsMultiselect',
         type:  'checkbox'
       },
+      {
+        title: browser.i18n.getMessage('config_loggingQueries_label'),
+        key:   'loggingQueries',
+        type:  'checkbox'
+      },
       { type: 'separator' },
       {
         title: browser.i18n.getMessage('config_debug_label'),
@@ -785,6 +1007,8 @@ function createItem(id, item, parent) {
     contexts: ['browser_action'],
     parentId
   };
+  if ('enabled' in item)
+    params.enabled = item.enabled;
   log('create: ', params);
   browser.menus.create(params);
   if (item.children) {
@@ -818,12 +1042,12 @@ browser.menus.onShown.addListener((info, _tab) => {
     if ('visible' in item)
       params.visible = item.visible;
     if ('checked' in params || 'title' in params) {
-      browser.menus.update(item.id, params);
+      browser.menus.update(item.id, params).catch(ApiTabs.createErrorSuppressor());
       updated = true;
     }
   }
   if (updated)
-    browser.menus.refresh();
+    browser.menus.refresh().catch(ApiTabs.createErrorSuppressor());
 });
 
 browser.menus.onClicked.addListener((info, _tab) => {
