@@ -10,12 +10,15 @@ import {
   configs
 } from '/common/common.js';
 
-import * as Tabs from '/common/tabs.js';
-import * as TabsGroup from '/common/tabs-group.js';
-import * as Commands from '/common/commands.js';
+import * as ApiTabs from '/common/api-tabs.js';
 import * as TSTAPI from '/common/tst-api.js';
 import * as TabContextMenu from './tab-context-menu.js';
 
+import Tab from '/common/Tab.js';
+
+import * as TabsGroup from './tabs-group.js';
+import * as Commands from './commands.js';
+
 function log(...args) {
   internalLogger('background/context-menu', ...args);
 }
@@ -167,7 +170,7 @@ function initItems() {
 }
 
 function updateItem(id, params) {
-  browser.menus.update(id, params);
+  browser.menus.update(id, params).catch(ApiTabs.createErrorSuppressor());
   TabContextMenu.onExternalMessage({
     type:   TSTAPI.kCONTEXT_MENU_UPDATE,
     params: [id, params]
@@ -248,99 +251,17 @@ function updateItems() {
   return updated;
 }
 
-// for Firefox 62 or olders
-export function refreshItems() {
-  if (mNativeContextMenuAvailable)
+export const onClick = (info, tab) => {
+  // Extra context menu commands won't be available on the blank area of the tab bar.
+  if (!tab)
     return;
 
-  browser.menus.remove(kROOT_ITEM);
-  TabContextMenu.onExternalMessage({
-    type:   TSTAPI.kCONTEXT_MENU_REMOVE,
-    params: kROOT_ITEM
-  }, browser.runtime);
+  log('context menu item clicked: ', info, tab);
 
-  let separatorsCount = 0;
-  let normalItemAppeared = false;
-  const items = [];
-  const customItems = [];
-  for (const item of mContextMenuItems) {
-    let id = item.id;
-    browser.menus.remove(id);
-    TabContextMenu.onExternalMessage({
-      type:   TSTAPI.kCONTEXT_MENU_REMOVE,
-      params: id
-    }, browser.runtime);
-    if (item.type == 'separator') {
-      if (!normalItemAppeared)
-        continue;
-      normalItemAppeared = false;
-      id = `separator${separatorsCount++}`;
-    }
-    else {
-      if (item.hidden || !configs[`context_${id}`])
-        continue;
-      normalItemAppeared = true;
-    }
-    items.push({
-      id,
-      type:     item.type,
-      checked:  item.checked,
-      title:    item.titleWithoutAccesskey,
-      contexts: ['tab']
-    });
-    customItems.push({
-      type: TSTAPI.kCONTEXT_MENU_CREATE,
-      params: {
-        id,
-        type:     item.type,
-        checked:  item.checked,
-        title:    item.title,
-        contexts: ['tab']
-      }
-    });
-  }
+  const contextTab = Tab.get(tab.id);
+  const selectedTabs = contextTab.$TST.multiselected ? Tab.getSelectedTabs(contextTab.windowId) : [];
 
-  if (items.length > 0 && items[items.length - 1].type == 'separator') {
-    items.pop();
-    customItems.pop();
-  }
-  if (items.length == 0)
-    return;
-
-  const grouped = configs.emulateDefaultContextMenu && items.length > 1;
-  if (grouped) {
-    const manifest = browser.runtime.getManifest();
-    browser.menus.create({
-      id:       kROOT_ITEM,
-      type:     'normal',
-      contexts: ['tab'],
-      title:    manifest.name,
-      icons:    manifest.icons
-    });
-    TabContextMenu.onExternalMessage({
-      type: TSTAPI.kCONTEXT_MENU_CREATE,
-      params: {
-        id:       kROOT_ITEM,
-        contexts: ['tab'],
-        title:    manifest.name,
-        icons:    manifest.icons
-      }
-    }, browser.runtime);
-  }
-  for (let i = 0, maxi = items.length; i < maxi; i++) {
-    items[i].parentId = customItems[i].params.parentId = grouped ? kROOT_ITEM : null ;
-    browser.menus.create(items[i]);
-    TabContextMenu.onExternalMessage(customItems[i], browser.runtime);
-  }
-}
-
-export const onClick = (info, apiTab) => {
-  log('context menu item clicked: ', info, apiTab);
-
-  const contextTab = Tabs.getTabById(apiTab);
-  const selectedTabs = Tabs.isMultiselected(contextTab) ? Tabs.getSelectedTabs(contextTab) : [];
-
-  switch (info.menuItemId.replace(/^(?:grouped:|context_closeTabOptions_)/, '')) {
+  switch (info.menuItemId.replace(/^(?:grouped:|context_topLevel_)/, '')) {
     case 'reloadTree':
       Commands.reloadTree(contextTab);
       break;
@@ -362,13 +283,13 @@ export const onClick = (info, apiTab) => {
       Commands.collapseTree(contextTab);
       break;
     case 'collapseAll':
-      Commands.collapseAll(contextTab);
+      Commands.collapseAll(contextTab.windowId);
       break;
     case 'expandTree':
       Commands.expandTree(contextTab);
       break;
     case 'expandAll':
-      Commands.expandAll(contextTab);
+      Commands.expandAll(contextTab.windowId);
       break;
 
     case 'bookmarkTree':
@@ -387,14 +308,16 @@ export const onClick = (info, apiTab) => {
         Commands.collapseTree(contextTab);
       break;
     case 'pinnedTab': {
-      const tabs = Tabs.getPinnedTabs(contextTab);
+      const tabs = Tab.getPinnedTabs(contextTab.windowId);
       if (tabs.length > 0)
-        browser.tabs.update(tabs[0].apiTab.id, { active: true });
+        browser.tabs.update(tabs[0].id, { active: true })
+          .catch(ApiTabs.createErrorHandler(ApiTabs.handleMissingTabError));
     }; break;
     case 'unpinnedTab': {
-      const tabs = Tabs.getUnpinnedTabs(contextTab);
+      const tabs = Tab.getUnpinnedTabs(tab.windowId);
       if (tabs.length > 0)
-        browser.tabs.update(tabs[0].apiTab.id, { active: true });
+        browser.tabs.update(tabs[0].id, { active: true })
+          .catch(ApiTabs.createErrorHandler(ApiTabs.handleMissingTabError));
     }; break;
 
     default:
@@ -414,10 +337,10 @@ function onShown(info, tab) {
     updated = updateItems();
   }
 
-  tab = tab && Tabs.getTabById(tab.id);
-  const subtreeCollapsed = Tabs.isSubtreeCollapsed(tab);
-  const hasChild = Tabs.hasChildTabs(tab);
-  const multiselected = Tabs.isMultiselected(tab);
+  tab = tab && Tab.get(tab.id);
+  const subtreeCollapsed = tab && tab.$TST.subtreeCollapsed;
+  const hasChild = tab && tab.$TST.hasChild;
+  const multiselected = tab && tab.$TST.multiselected;
 
   for (const item of mContextMenuItems) {
     let newEnabled;
@@ -465,7 +388,7 @@ function onShown(info, tab) {
   }
 
   if (updated)
-    browser.menus.refresh();
+    browser.menus.refresh().catch(ApiTabs.createErrorSuppressor());
 }
 browser.menus.onShown.addListener(onShown);
 TabContextMenu.onTSTTabContextMenuShown.addListener(onShown);
