@@ -12,6 +12,7 @@
   let gTitle;
   let gTitleField;
   let gTemporaryCheck;
+  let gTemporaryAggressiveCheck;
 
   document.title = getTitle();
 
@@ -36,6 +37,11 @@
     return /[&?]temporary=true/.test(params);
   }
 
+  function isTemporaryAggressive() {
+    const params = location.search.split('#')[0];
+    return /[&?]temporaryAggressive=true/.test(params);
+  }
+
   function getOpenerTabId() {
     const params = location.search.split('#')[0];
     const matched = params.match(/[&?]openerTabId=([^&;]*)/);
@@ -66,14 +72,19 @@
   }
 
   function updateParameters(aParameters = {}) {
-    const title     = aParameters.title || getTitle() || '';
-    const temporary = String(gTemporaryCheck.checked);
+    const title = aParameters.title || getTitle() || '';
 
-    let opener    = getOpenerTabId();
+    let temporary = gTemporaryCheck.checked;
+    temporary = temporary ? `&temporary=${temporary}` : '';
+
+    let temporaryAggressive = gTemporaryAggressiveCheck.checked;
+    temporaryAggressive = temporaryAggressive ? `&temporaryAggressive=${temporaryAggressive}` : '';
+
+    let opener = getOpenerTabId();
     opener = opener ? `&openerTabId=${opener}` : '';
 
     let uri = location.href.split('?')[0];
-    uri = `${uri}?title=${encodeURIComponent(title)}&temporary=${temporary}${opener}`;
+    uri = `${uri}?title=${encodeURIComponent(title)}${temporary}${temporaryAggressive}${opener}`;
     history.replaceState({}, document.title, uri);
   }
 
@@ -140,11 +151,82 @@
 
     gTemporaryCheck = document.querySelector('#temporary');
     gTemporaryCheck.checked = isTemporary();
-    gTemporaryCheck.addEventListener('change', _event => updateParameters());
+    gTemporaryCheck.addEventListener('change', _event => {
+      if (gTemporaryCheck.checked)
+        gTemporaryAggressiveCheck.checked = false;
+      updateParameters();
+    });
+
+    gTemporaryAggressiveCheck = document.querySelector('#temporaryAggressive');
+    gTemporaryAggressiveCheck.checked = isTemporaryAggressive();
+    gTemporaryAggressiveCheck.addEventListener('change', _event => {
+      if (gTemporaryAggressiveCheck.checked)
+        gTemporaryCheck.checked = false;
+      updateParameters();
+    });
 
     window.l10n.updateDocument();
 
-    updateTree();
+    browser.runtime.sendMessage({
+      type: 'treestyletab:get-config-value',
+      keys: [
+        'renderTreeInGroupTabs',
+        'showAutoGroupOptionHint'
+      ]
+    }).then(configs => {
+      updateTree.enabled = configs.renderTreeInGroupTabs;
+      updateTree();
+
+      let show = configs.showAutoGroupOptionHint;
+      if (!isTemporary() && !isTemporaryAggressive())
+        show = false;
+
+      const hint = document.getElementById('optionHint');
+      hint.style.display = show ? 'block' : 'none';
+      if (!show)
+        return;
+
+      hint.firstChild.addEventListener('click', event => {
+        if (event.button != 0)
+          return;
+        browser.runtime.sendMessage({
+          type: 'treestyletab:open-tab',
+          uri:  `moz-extension://${location.host}/options/options.html#autoGroupNewTabsSection`
+        });
+      });
+      hint.firstChild.addEventListener('keydown', event => {
+        if (event.key != 'Enter' &&
+            event.key != 'Space')
+          return;
+        browser.runtime.sendMessage({
+          type: 'treestyletab:open-tab',
+          uri:  `moz-extension://${location.host}/options/options.html#autoGroupNewTabsSection`
+        });
+      });
+
+      const closebox = document.getElementById('dismissOptionHint');
+      closebox.addEventListener('click', event => {
+        if (event.button != 0)
+          return;
+        hint.style.display = 'none';
+        browser.runtime.sendMessage({
+          type: 'treestyletab:set-config-value',
+          key:  'showAutoGroupOptionHint',
+          value: false
+        });
+      });
+      closebox.addEventListener('keydown', event => {
+        if (event.key != 'Enter' &&
+            event.key != 'Space')
+          return;
+        hint.style.display = 'none';
+        browser.runtime.sendMessage({
+          type: 'treestyletab:set-config-value',
+          key:  'showAutoGroupOptionHint',
+          value: false
+        });
+      });
+    });
 
     window.setTitle    = window.setTitle || setTitle;
     window.updateTree  = window.updateTree || updateTree;
@@ -154,30 +236,47 @@
 
 
   async function updateTree() {
-    const tabs = await browser.runtime.sendMessage({
-      type: 'treestyletab:api:get-tree',
-      tabs: [
-        'senderTab',
-        getOpenerTabId()
-      ]
-    });
+    const runAt = updateTree.lastRun = Date.now();
+
     const container = document.getElementById('tabs');
     const range = document.createRange();
     range.selectNodeContents(container);
     range.deleteContents();
     range.detach();
+
+    if (!updateTree.enabled)
+      return;
+
+    document.documentElement.classList.add('updating');
+
+    const tabs = await browser.runtime.sendMessage({
+      type: 'treestyletab:api:get-tree',
+      tabs: [
+        'senderTab',
+        getOpenerTabId()
+      ],
+      interval: 50
+    });
+
+    // called again while waiting
+    if (runAt != updateTree.lastRun)
+      return;
+
     let tree;
     if (tabs[1]) {
       tabs[1].children = tabs[0].children;
-      tree = buildTabChildren({ children: [tabs[1]] });
+      tree = buildChildren({ children: [tabs[1]] });
     }
     else
-      tree = buildTabChildren(tabs[0]);
+      tree = buildChildren(tabs[0]);
     if (tree) {
       container.appendChild(tree);
       reflow();
     }
+
+    document.documentElement.classList.remove('updating');
   }
+  updateTree.enabled = true;
 
   function reflow() {
     const container = document.getElementById('tabs');
@@ -187,7 +286,7 @@
     });
   }
 
-  function buildTabItem(tab) {
+  function buildItem(tab) {
     const item = document.createElement('li');
 
     const link = item.appendChild(document.createElement('a'));
@@ -215,7 +314,7 @@
     label.classList.add('label');
     label.textContent = tab.title;
 
-    const children = buildTabChildren(tab);
+    const children = buildChildren(tab);
     if (!children)
       return item;
 
@@ -227,11 +326,11 @@
     return fragment;
   }
 
-  function buildTabChildren(tab) {
+  function buildChildren(tab) {
     if (tab.children && tab.children.length > 0) {
       const list = document.createElement('ul');
       for (const child of tab.children) {
-        list.appendChild(buildTabItem(child));
+        list.appendChild(buildItem(child));
       }
       return list;
     }
