@@ -13,10 +13,13 @@ import {
 
 import * as Constants from '/common/constants.js';
 import * as ApiTabs from '/common/api-tabs.js';
-import * as Tabs from '/common/tabs.js';
-import * as Tree from '/common/tree.js';
-import * as Commands from '/common/commands.js';
+import * as TabsStore from '/common/tabs-store.js';
+import * as TreeBehavior from '/common/tree-behavior.js';
 
+import Tab from '/common/Tab.js';
+
+import * as Commands from './commands.js';
+import * as Tree from './tree.js';
 import * as TreeStructure from './tree-structure.js';
 
 function log(...args) {
@@ -27,55 +30,58 @@ function logApiTabs(...args) {
 }
 
 
-Tabs.onCreated.addListener((tab, info = {}) => {
-  if (info.duplicated ||
-      info.restored ||
-      info.skipFixupTree ||
-      // do nothing for already attached tabs
-      (tab.apiTab.openerTabId &&
-       Tabs.getParentTab(tab) == Tabs.getTabById(tab.apiTab.openerTabId)))
+Tab.onCreated.addListener((tab, info = {}) => {
+  if (!info.mayBeReplacedWithContainer &&
+      (info.duplicated ||
+       info.restored ||
+       info.skipFixupTree ||
+       // do nothing for already attached tabs
+       (tab.openerTabId &&
+        tab.$TST.parent == Tab.get(tab.openerTabId))))
     return;
   // if the tab is opened inside existing tree by someone, we must fixup the tree.
-  if (!info.positionedBySelf &&
-      (Tabs.getNextNormalTab(tab) ||
-       Tabs.getPreviousNormalTab(tab) ||
+  if (!(info.positionedBySelf ||
+        info.movedBySelfWhileCreation) &&
+      (tab.$TST.nearestCompletelyOpenedNormalFollowingTab ||
+       tab.$TST.nearestCompletelyOpenedNormalPrecedingTab ||
        (info.treeForActionDetection &&
+        info.treeForActionDetection.target &&
         (info.treeForActionDetection.target.next ||
          info.treeForActionDetection.target.previous))))
     tryFixupTreeForInsertedTab(tab, {
-      toIndex:   tab.apiTab.index,
-      fromIndex: Tabs.getTabIndex(Tabs.getLastTab(tab)),
+      toIndex:   tab.index,
+      fromIndex: Tab.getLastTab(tab.windowId).index,
       treeForActionDetection: info.treeForActionDetection
     });
 });
 
-Tabs.onMoving.addListener((tab, moveInfo) => {
+Tab.onMoving.addListener((tab, moveInfo) => {
   // avoid TabMove produced by browser.tabs.insertRelatedAfterCurrent=true or something.
-  const container = Tabs.getTabsContainer(tab);
-  const isNewlyOpenedTab = container.openingTabs.has(tab.apiTab.id);
+  const window           = TabsStore.windows.get(tab.windowId);
+  const isNewlyOpenedTab = window.openingTabs.has(tab.id);
   const positionControlled = configs.insertNewChildAt != Constants.kINSERT_NO_CONTROL;
-  if (!isNewlyOpenedTab ||
-      moveInfo.byInternalOperation ||
-      moveInfo.alreadyMoved ||
-      !positionControlled)
-    return true;
-
-  const opener = Tabs.getOpenerTab(tab);
-  // if there is no valid opener, it can be a restored initial tab in a restored window
-  // and can be just moved as a part of window restoration process.
-  if (!opener)
-    return true;
-
-  log('onTabMove for new child tab: move back '+moveInfo.toIndex+' => '+moveInfo.fromIndex);
-  moveBack(tab, moveInfo);
-  return false;
+  if (isNewlyOpenedTab &&
+      !moveInfo.byInternalOperation &&
+      !moveInfo.alreadyMoved &&
+      !moveInfo.isSubstantiallyMoved &&
+      positionControlled) {
+    const opener = tab.$TST.openerTab;
+    // if there is no valid opener, it can be a restored initial tab in a restored window
+    // and can be just moved as a part of window restoration process.
+    if (opener) {
+      log('onTabMove for new child tab: move back '+moveInfo.toIndex+' => '+moveInfo.fromIndex);
+      moveBack(tab, moveInfo);
+      return false;
+    }
+  }
+  return true;
 });
 
 async function tryFixupTreeForInsertedTab(tab, moveInfo = {}) {
-  if (!Tree.shouldApplyTreeBehavior(moveInfo)) {
+  if (!TreeBehavior.shouldApplyTreeBehavior(moveInfo)) {
     Tree.detachAllChildren(tab, {
-      behavior: Tree.getCloseParentBehaviorForTab(tab, {
-        keepChildren: true
+      behavior: TreeBehavior.getCloseParentBehaviorForTab(tab, {
+        applyTreeBehavior: true
       }),
       broadcast: true
     });
@@ -85,7 +91,7 @@ async function tryFixupTreeForInsertedTab(tab, moveInfo = {}) {
   }
 
   log('the tab can be placed inside existing tab unexpectedly, so now we are trying to fixup tree.');
-  const action = await detectTabActionFromNewPosition(tab, moveInfo);
+  const action = detectTabActionFromNewPosition(tab, moveInfo);
   if (!action) {
     log('no action');
     return;
@@ -98,9 +104,9 @@ async function tryFixupTreeForInsertedTab(tab, moveInfo = {}) {
       return;
 
     case 'attach': {
-      await Tree.attachTabTo(tab, Tabs.getTabById(action.parent), {
-        insertBefore: Tabs.getTabById(action.insertBefore),
-        insertAfter:  Tabs.getTabById(action.insertAfter),
+      await Tree.attachTabTo(tab, Tab.get(action.parent), {
+        insertBefore: Tab.get(action.insertBefore),
+        insertAfter:  Tab.get(action.insertAfter),
         broadcast:    true
       });
       Tree.followDescendantsToMovedRoot(tab);
@@ -117,30 +123,29 @@ async function tryFixupTreeForInsertedTab(tab, moveInfo = {}) {
   }
 }
 
-Tabs.onMoved.addListener((tab, moveInfo = {}) => {
-  if (moveInfo.byInternalOperation ||
-      Tabs.isDuplicating(tab)) {
+Tab.onMoved.addListener((tab, moveInfo = {}) => {
+  if (!moveInfo.byInternalOperation &&
+      !moveInfo.isSubstantiallyMoved &&
+      !tab.$TST.duplicating) {
+    log('process moved tab');
+    tryFixupTreeForInsertedTab(tab, moveInfo);
+  }
+  else {
     log('internal move');
-    return;
   }
-  log('process moved tab');
-
-  tryFixupTreeForInsertedTab(tab, moveInfo);
 });
 
 Commands.onMoveUp.addListener(async tab => {
-  const index = Tabs.getTabIndex(tab);
   await tryFixupTreeForInsertedTab(tab, {
-    toIndex:   index,
-    fromIndex: index + 1,
+    toIndex:   tab.index,
+    fromIndex: tab.index + 1,
   });
 });
 
 Commands.onMoveDown.addListener(async tab => {
-  const index = Tabs.getTabIndex(tab);
   await tryFixupTreeForInsertedTab(tab, {
-    toIndex:   index,
-    fromIndex: index - 1,
+    toIndex:   tab.index,
+    fromIndex: tab.index - 1,
   });
 });
 
@@ -148,27 +153,31 @@ TreeStructure.onTabAttachedFromRestoredInfo.addListener((tab, moveInfo) => { try
 
 function moveBack(tab, moveInfo) {
   log('Move back tab from unexpected move: ', dumpTab(tab), moveInfo);
-  const id = tab.apiTab.id;
-  const container = tab.parentNode;
-  container.internalMovingTabs.add(id);
-  logApiTabs(`handle-moved-tabs:moveBack: browser.tabs.move() `, tab.apiTab.id, {
+  const id     = tab.id;
+  const window = TabsStore.windows.get(tab.windowId);
+  window.internalMovingTabs.add(id);
+  logApiTabs(`handle-moved-tabs:moveBack: browser.tabs.move() `, tab.id, {
     windowId: moveInfo.windowId,
     index:    moveInfo.fromIndex
   });
-  return browser.tabs.move(tab.apiTab.id, {
+  return browser.tabs.move(tab.id, {
     windowId: moveInfo.windowId,
     index:    moveInfo.fromIndex
-  }).catch(e => {
-    if (container.internalMovingTabs.has(id))
-      container.internalMovingTabs.delete(id);
+  }).catch(ApiTabs.createErrorHandler(e => {
+    if (window.internalMovingTabs.has(id))
+      window.internalMovingTabs.delete(id);
     ApiTabs.handleMissingTabError(e);
-  });
+  }));
 }
 
-async function detectTabActionFromNewPosition(tab, moveInfo = {}) {
+function detectTabActionFromNewPosition(tab, moveInfo = {}) {
+  if (tab.pinned)
+    return tab.$TST.parentId ? { action: 'detach' } : { action: 'move' };
+
   log('detectTabActionFromNewPosition: ', dumpTab(tab), moveInfo);
-  const tree   = moveInfo.treeForActionDetection || Tabs.snapshotTreeForActionDetection(tab);
+  const tree   = moveInfo.treeForActionDetection || Tree.snapshotForActionDetection(tab);
   const target = tree.target;
+  log('  calculate new position: ', tab, tree);
 
   const toIndex   = moveInfo.toIndex;
   const fromIndex = moveInfo.fromIndex;
@@ -179,8 +188,8 @@ async function detectTabActionFromNewPosition(tab, moveInfo = {}) {
 
   const prevTab = tree.tabsById[target.previous];
   const nextTab = tree.tabsById[target.next];
-  log('prevTab: ', prevTab && prevTab.id);
-  log('nextTab: ', nextTab && nextTab.id);
+  log('prevTab: ', dumpTab(prevTab));
+  log('nextTab: ', dumpTab(nextTab));
 
   const prevParent = prevTab && tree.tabsById[prevTab.parent];
   const nextParent = nextTab && tree.tabsById[nextTab.parent];
@@ -193,12 +202,9 @@ async function detectTabActionFromNewPosition(tab, moveInfo = {}) {
   const oldParent = tree.tabsById[target.parent];
   let newParent = null;
 
-  if (prevTab &&
-      target.cookieStoreId != prevTab.cookieStoreId &&
-      target.url == prevTab.url) {
-    // https://addons.mozilla.org/en-US/firefox/addon/multi-account-containers/
-    log('=> replaced by Firefox Multi-Acount Containers');
-    newParent = prevParent;
+  if (target.mayBeReplacedWithContainer) {
+    log('=> replaced by Firefox Multi-Acount Containers or Temporary Containers');
+    newParent = prevLevel < nextLevel ? prevTab : prevParent;
   }
   else if (oldParent &&
            prevTab &&
@@ -231,7 +237,9 @@ async function detectTabActionFromNewPosition(tab, moveInfo = {}) {
   }
   else if (prevLevel > nextLevel) {
     log('=> moved to end of existing tree');
-    if (!target.active) {
+    if (!target.active &&
+        target.children.length == 0 &&
+        (Date.now() - target.trackedAt) < 500) {
       log('=> maybe newly opened tab');
       newParent = prevParent;
     }
