@@ -12,11 +12,15 @@ import {
 } from '/common/common.js';
 
 import * as Constants from '/common/constants.js';
-import * as Tabs from '/common/tabs.js';
+import * as TabsStore from '/common/tabs-store.js';
 import * as TabsInternalOperation from '/common/tabs-internal-operation.js';
-import * as TabsMove from '/common/tabs-move.js';
-import * as TabsOpen from '/common/tabs-open.js';
-import * as Tree from '/common/tree.js';
+import * as TreeBehavior from '/common/tree-behavior.js';
+
+import Tab from '/common/Tab.js';
+
+import * as TabsOpen from './tabs-open.js';
+import * as TabsMove from './tabs-move.js';
+import * as Tree from './tree.js';
 
 function log(...args) {
   internalLogger('background/handle-new-tabs', ...args);
@@ -24,55 +28,80 @@ function log(...args) {
 
 
 // this should return false if the tab is / may be moved while processing
-Tabs.onCreating.addListener((tab, info = {}) => {
+Tab.onCreating.addListener((tab, info = {}) => {
   if (info.duplicatedInternally)
     return true;
 
   log('Tabs.onCreating ', dumpTab(tab), info);
 
-  const possibleOpenerTab = info.activeTab || Tabs.getActiveTab(tab);
-  const opener = Tabs.getOpenerTab(tab);
-  if (opener)
-    opener.uniqueId.then(uniqueId => {
-      tab.dataset.originalOpenerTabId = uniqueId.id;
-    });
-
-  if (!opener) {
+  const possibleOpenerTab = info.activeTab || Tab.getActiveTab(tab.windowId);
+  const opener = tab.$TST.openerTab;
+  if (opener) {
+    tab.$TST.setAttribute(Constants.kPERSISTENT_ORIGINAL_OPENER_TAB_ID, opener.$TST.uniqueId.id);
+    TabsStore.addToBeGroupedTab(tab);
+  }
+  else {
     if (!info.maybeOrphan &&
-        possibleOpenerTab &&
-        /* New tab opened with browser.tabs.insertAfterCurrent=true may have
-           next tab. In this case the tab is expected to be placed next to the
-           active tab aways, so we should skip all repositioning behavior.
-           See also: https://github.com/piroor/treestyletab/issues/2054 */
-        !Tabs.getNextTab(tab)) {
-      if (Tabs.isNewTabCommandTab(tab)) {
+        possibleOpenerTab) {
+      let autoAttachBehavior = configs.autoAttachOnNewTabCommand;
+      let dontMove           = false;
+      if (tab.$TST.nextTab &&
+          possibleOpenerTab == tab.$TST.previousTab) {
+        // New tab opened with browser.tabs.insertAfterCurrent=true may have
+        // next tab. In this case the tab is expected to be placed next to the
+        // active tab always, so we should change the behavior specially.
+        // See also:
+        //   https://github.com/piroor/treestyletab/issues/2054
+        //   https://github.com/piroor/treestyletab/issues/2194#issuecomment-505272940
+        dontMove = true;
+        switch (autoAttachBehavior) {
+          case Constants.kNEWTAB_OPEN_AS_ORPHAN:
+          case Constants.kNEWTAB_OPEN_AS_SIBLING:
+          case Constants.kNEWTAB_OPEN_AS_NEXT_SIBLING:
+            if (possibleOpenerTab.$TST.hasChild)
+              autoAttachBehavior = Constants.kNEWTAB_OPEN_AS_CHILD;
+            else
+              autoAttachBehavior = Constants.kNEWTAB_OPEN_AS_NEXT_SIBLING;
+            break;
+
+          case Constants.kNEWTAB_OPEN_AS_CHILD:
+          default:
+            break;
+        }
+      }
+      if (tab.$TST.isNewTabCommandTab) {
         if (!info.positionedBySelf) {
           log('behave as a tab opened by new tab command');
           return handleNewTabFromActiveTab(tab, {
             activeTab:                 possibleOpenerTab,
-            autoAttachBehavior:        configs.autoAttachOnNewTabCommand,
+            autoAttachBehavior,
+            dontMove,
             inheritContextualIdentity: configs.inheritContextualIdentityToNewChildTab
           }).then(moved => !moved);
         }
         return false;
       }
       else if (possibleOpenerTab != tab) {
-        tab.dataset.possibleOpenerTab = possibleOpenerTab.id;
+        tab.$TST.possibleOpenerTab = possibleOpenerTab.id;
       }
-      tab.dataset.isNewTab = true;
+      tab.$TST.isNewTab = true;
     }
     log('behave as a tab opened with any URL');
+    tab.$TST.positionedBySelf = info.positionedBySelf;
     return true;
   }
 
   log(`opener: ${dumpTab(opener)}, positionedBySelf = ${info.positionedBySelf}`);
-  if (Tabs.isPinned(opener) &&
-      opener.parentNode == tab.parentNode) {
+  if (opener && opener.pinned &&
+      opener.windowId == tab.windowId) {
     if (configs.autoGroupNewTabsFromPinned) {
-      return false;
+      if (Tab.getGroupTabForOpener(opener) ||
+          tab.$TST.needToBeGroupedSiblings.length > 0)
+        return false;
+      return true;
     }
     if (configs.insertNewTabFromPinnedTabAt == Constants.kINSERT_END) {
-      return TabsMove.moveTabAfter(tab, Tabs.getLastTab(tab), {
+      return TabsMove.moveTabAfter(tab, Tab.getLastTab(tab.windowId), {
         delayedMove: true,
         broadcast:   true
       }).then(moved => !moved);
@@ -82,7 +111,7 @@ Tabs.onCreating.addListener((tab, info = {}) => {
     return Tree.behaveAutoAttachedTab(tab, {
       baseTab:   opener,
       behavior:  configs.autoAttachOnOpenedWithOwner,
-      dontMove:  info.positionedBySelf,
+      dontMove:  info.positionedBySelf || info.mayBeReplacedWithContainer,
       broadcast: true
     }).then(moved => !moved);
   }
@@ -92,20 +121,31 @@ Tabs.onCreating.addListener((tab, info = {}) => {
 async function handleNewTabFromActiveTab(tab, params = {}) {
   const activeTab = params.activeTab;
   log('handleNewTabFromActiveTab: activeTab = ', dumpTab(activeTab), params);
+  if (activeTab &&
+      activeTab.$TST.ancestors.includes(tab)) {
+    log(' => ignore restored ancestor tab');
+    return;
+  }
   const moved = await Tree.behaveAutoAttachedTab(tab, {
     baseTab:   activeTab,
     behavior:  params.autoAttachBehavior,
-    broadcast: true
+    broadcast: true,
+    dontMove:  params.dontMove || false
   });
-  const parent = Tabs.getParentTab(tab);
+  const parent = tab.$TST.parent;
   if (!parent ||
       !params.inheritContextualIdentity ||
-      tab.apiTab.cookieStoreId != 'firefox-default' ||
-      tab.apiTab.cookieStoreId == parent.apiTab.cookieStoreId)
+      tab.cookieStoreId != 'firefox-default' ||
+      tab.cookieStoreId == parent.cookieStoreId)
     return moved;
-  const cookieStoreId = activeTab.apiTab.cookieStoreId;
+  const cookieStoreId = activeTab.cookieStoreId;
   log('handleNewTabFromActiveTab: reopen with inherited contextual identity ', cookieStoreId);
+  // We need to prevent grouping of this original tab and the reopened tab
+  // by the "multiple tab opened in XXX msec" feature.
+  const window = TabsStore.windows.get(tab.windowId);
+  window.openedNewTabs.delete(tab.id);
   await TabsOpen.openNewTab({
+    windowId: activeTab.windowId,
     parent,
     insertBefore: tab,
     cookieStoreId
@@ -114,59 +154,59 @@ async function handleNewTabFromActiveTab(tab, params = {}) {
   return moved;
 }
 
-Tabs.onCreated.addListener((tab, info = {}) => {
+Tab.onCreated.addListener((tab, info = {}) => {
   if (!info.duplicated)
     return;
   const original = info.originalTab;
   log('duplicated ', dumpTab(tab), dumpTab(original));
   if (info.duplicatedInternally) {
     log('duplicated by internal operation');
-    tab.classList.add(Constants.kTAB_STATE_DUPLICATING);
-    Tabs.broadcastTabState(tab, {
-      add: [Constants.kTAB_STATE_DUPLICATING]
-    });
+    tab.$TST.addState(Constants.kTAB_STATE_DUPLICATING, { broadcast: true });
+    TabsStore.addDuplicatingTab(tab);
   }
   else {
     Tree.behaveAutoAttachedTab(tab, {
       baseTab:   original,
       behavior:  configs.autoAttachOnDuplicated,
-      dontMove:  info.positionedBySelf,
+      dontMove:  info.positionedBySelf || info.movedBySelfWhileCreation || info.mayBeReplacedWithContainer,
       broadcast: true
     });
   }
 });
 
-Tabs.onUpdated.addListener((tab, changeInfo) => {
-  if ('openerTabId' in changeInfo && configs.syncParentTabAndOpenerTab) {
-    Tabs.waitUntilAllTabsAreCreated(tab.apiTab.windowId).then(() => {
-      const parent = Tabs.getOpenerTab(tab);
+Tab.onUpdated.addListener((tab, changeInfo) => {
+  if ('openerTabId' in changeInfo &&
+      configs.syncParentTabAndOpenerTab &&
+      !tab.$TST.updatingOpenerTabIds.includes(changeInfo.openerTabId) /* accept only changes from outside of TST */) {
+    Tab.waitUntilTrackedAll(tab.windowId).then(() => {
+      const parent = tab.$TST.openerTab;
       if (!parent ||
-          parent.parentNode != tab.parentNode ||
-          parent == Tabs.getParentTab(tab))
+          parent.windowId != tab.windowId ||
+          parent == tab.$TST.parent)
         return;
       Tree.attachTabTo(tab, parent, {
         insertAt:    Constants.kINSERT_NEAREST,
-        forceExpand: Tabs.isActive(tab),
+        forceExpand: tab.active,
         broadcast:   true
       });
     });
   }
 
   if ((changeInfo.url || changeInfo.status == 'complete') &&
-      tab.dataset.isNewTab) {
+      tab.$TST.isNewTab) {
     log('new tab ', dumpTab(tab));
-    delete tab.dataset.isNewTab;
-    const possibleOpenerTab = Tabs.getTabById(tab.dataset.possibleOpenerTab);
-    delete tab.dataset.possibleOpenerTab;
+    delete tab.$TST.isNewTab;
+    const possibleOpenerTab = Tab.get(tab.$TST.possibleOpenerTab);
+    delete tab.$TST.possibleOpenerTab;
     log('possibleOpenerTab ', dumpTab(possibleOpenerTab));
-    const toBeGroupedTabs = tab.parentNode.openedNewTabs
-      .map(Tabs.getTabById)
-      .filter(tab => !!tab);
-    log('toBeGroupedTabs ', toBeGroupedTabs.map(dumpTab));
-    if (!Tabs.getParentTab(tab) &&
+    const window = TabsStore.windows.get(tab.windowId);
+    log('window.openedNewTabs ', window.openedNewTabs);
+    if (!tab.$TST.parent &&
         possibleOpenerTab &&
-        !toBeGroupedTabs.includes(tab)) {
-      if (Tabs.isNewTabCommandTab(tab)) {
+        !window.openedNewTabs.has(tab.id) &&
+        !tab.$TST.openedWithOthers &&
+        !tab.$TST.positionedBySelf) {
+      if (tab.$TST.isNewTabCommandTab) {
         log('behave as a tab opened by new tab command (delayed)');
         handleNewTabFromActiveTab(tab, {
           activeTab:                 possibleOpenerTab,
@@ -176,8 +216,8 @@ Tabs.onUpdated.addListener((tab, changeInfo) => {
       }
       else {
         const siteMatcher  = /^\w+:\/\/([^\/]+)(?:$|\/.*$)/;
-        const openerTabSite = possibleOpenerTab.apiTab.url.match(siteMatcher);
-        const newTabSite    = tab.apiTab.url.match(siteMatcher);
+        const openerTabSite = possibleOpenerTab.url.match(siteMatcher);
+        const newTabSite    = tab.url.match(siteMatcher);
         if (openerTabSite && newTabSite && openerTabSite[1] == newTabSite[1]) {
           log('behave as a tab opened from same site (delayed)');
           handleNewTabFromActiveTab(tab, {
@@ -192,16 +232,16 @@ Tabs.onUpdated.addListener((tab, changeInfo) => {
 });
 
 
-Tabs.onAttached.addListener(async (tab, info = {}) => {
+Tab.onAttached.addListener(async (tab, info = {}) => {
   if (!info.windowId ||
-      !Tree.shouldApplyTreeBehavior(info))
+      !TreeBehavior.shouldApplyTreeBehavior(info))
     return;
 
   log('Tabs.onAttached ', dumpTab(tab), info);
 
   log('descendants of attached tab: ', info.descendants.map(dumpTab));
   const movedTabs = await Tree.moveTabs(info.descendants, {
-    destinationWindowId: tab.apiTab.windowId,
+    destinationWindowId: tab.windowId,
     insertAfter:         tab
   });
   log('moved descendants: ', movedTabs.map(dumpTab));
