@@ -13,67 +13,64 @@ import {
 } from '/common/common.js';
 
 import * as Constants from '/common/constants.js';
-import * as Tabs from '/common/tabs.js';
+import * as ApiTabs from '/common/api-tabs.js';
+import * as TabsStore from '/common/tabs-store.js';
 import * as TabsInternalOperation from '/common/tabs-internal-operation.js';
-import * as TabsOpen from '/common/tabs-open.js';
-import * as TabsGroup from '/common/tabs-group.js';
-import * as Tree from '/common/tree.js';
-import * as SidebarStatus from '/common/sidebar-status.js';
+import * as TreeBehavior from '/common/tree-behavior.js';
+import * as SidebarConnection from '/common/sidebar-connection.js';
+
+import Tab from '/common/Tab.js';
 
 import * as Background from './background.js';
+import * as TabsGroup from './tabs-group.js';
+import * as TabsOpen from './tabs-open.js';
+import * as Tree from './tree.js';
 
 function log(...args) {
   internalLogger('background/handle-removed-tabs', ...args);
 }
 
 
-Tabs.onRemoving.addListener(async (tab, removeInfo = {}) => {
-  log('Tabs.onRemoving ', dumpTab(tab), tab.apiTab, removeInfo);
+Tab.onRemoving.addListener(async (tab, removeInfo = {}) => {
+  log('Tabs.onRemoving ', dumpTab(tab), removeInfo);
   if (removeInfo.isWindowClosing)
     return;
 
-  let closeParentBehavior = Tree.getCloseParentBehaviorForTabWithSidebarOpenState(tab, removeInfo);
-  if (!SidebarStatus.isOpen(tab.apiTab.windowId) &&
+  let closeParentBehavior;
+  if (tab.$TST.mayBeOriginalOfReplacedWithContainer)
+    closeParentBehavior = Constants.kCLOSE_PARENT_BEHAVIOR_PROMOTE_FIRST_CHILD;
+  else
+    closeParentBehavior = TreeBehavior.getCloseParentBehaviorForTabWithSidebarOpenState(tab, removeInfo);
+
+  if (!SidebarConnection.isOpen(tab.windowId) &&
       closeParentBehavior != Constants.kCLOSE_PARENT_BEHAVIOR_CLOSE_ALL_CHILDREN &&
-      Tabs.isSubtreeCollapsed(tab))
+      tab.$TST.subtreeCollapsed)
     Tree.collapseExpandSubtree(tab, {
       collapsed: false,
       justNow:   true,
       broadcast: false // because the tab is going to be closed, broadcasted Tree.collapseExpandSubtree can be ignored.
     });
 
-  const wasActive = Tabs.isActive(tab);
   if (!(await tryGrantCloseTab(tab, closeParentBehavior)))
     return;
   log('Tabs.onRemoving: granted to close ', dumpTab(tab));
 
-  if (typeof browser.tabs.moveInSuccession != 'function') { // on Firefox 64 or older
-    const nextTab = closeParentBehavior == Constants.kCLOSE_PARENT_BEHAVIOR_CLOSE_ALL_CHILDREN && Tabs.getNextSiblingTab(tab) || Tabs.getNextTab(tab);
-    Tree.tryMoveFocusFromClosingActiveTab(tab, {
-      wasActive,
-      params: {
-        active:          wasActive,
-        nextTab:         nextTab && nextTab.id,
-        nextTabUrl:      nextTab && nextTab.apiTab.url,
-        nextIsDiscarded: Tabs.isDiscarded(nextTab)
-      }
-    });
-  }
-
   if (closeParentBehavior == Constants.kCLOSE_PARENT_BEHAVIOR_CLOSE_ALL_CHILDREN)
     await closeChildTabs(tab);
 
   if (closeParentBehavior == Constants.kCLOSE_PARENT_BEHAVIOR_REPLACE_WITH_GROUP_TAB &&
-      Tabs.getChildTabs(tab).length > 1) {
+      tab.$TST.childIds.length > 1 &&
+      tab.$TST.children.filter(child => !child.$TST.states.has(Constants.kTAB_STATE_TO_BE_REMOVED)).length > 1) {
     log('trying to replace the closing tab with a new group tab');
-    const firstChild = Tabs.getFirstChildTab(tab);
+    const firstChild = tab.$TST.firstChild;
     const uri = TabsGroup.makeGroupTabURI({
-      title:     browser.i18n.getMessage('groupTab_label', firstChild.apiTab.title),
-      temporary: true
+      title:     browser.i18n.getMessage('groupTab_label', firstChild.title),
+      temporaryAggressive: true
     });
-    tab.parentNode.toBeOpenedTabsWithPositions++;
+    const window = TabsStore.windows.get(tab.windowId);
+    window.toBeOpenedTabsWithPositions++;
     const groupTab = await TabsOpen.openURIInTab(uri, {
-      windowId:     tab.apiTab.windowId,
+      windowId:     tab.windowId,
       insertBefore: tab, // not firstChild, because the "tab" is disappeared from tree.
       inBackground: true
     });
@@ -85,6 +82,10 @@ Tabs.onRemoving.addListener(async (tab, removeInfo = {}) => {
       broadcast:    true
     });
     closeParentBehavior = Constants.kCLOSE_PARENT_BEHAVIOR_PROMOTE_FIRST_CHILD;
+    // This can be triggered on closing of multiple tabs,
+    // so we should cleanup it on such cases for safety.
+    // https://github.com/piroor/treestyletab/issues/2317
+    wait(1000).then(() => TabsGroup.reserveToCleanupNeedlessGroupTab(groupTab));
   }
 
   Tree.detachAllChildren(tab, {
@@ -99,9 +100,9 @@ Tabs.onRemoving.addListener(async (tab, removeInfo = {}) => {
 });
 
 async function tryGrantCloseTab(tab, closeParentBehavior) {
-  log('tryGrantClose: ', { alreadyGranted: configs.grantedRemovingTabIds, closing: tab.apiTab.id });
-  const alreadyGranted = configs.grantedRemovingTabIds.includes(tab.apiTab.id);
-  configs.grantedRemovingTabIds = configs.grantedRemovingTabIds.filter(id => id != tab.apiTab.id);
+  log('tryGrantClose: ', { alreadyGranted: configs.grantedRemovingTabIds, closing: dumpTab(tab) });
+  const alreadyGranted = configs.grantedRemovingTabIds.includes(tab.id);
+  configs.grantedRemovingTabIds = configs.grantedRemovingTabIds.filter(id => id != tab.id);
   if (alreadyGranted)
     return true;
 
@@ -110,54 +111,64 @@ async function tryGrantCloseTab(tab, closeParentBehavior) {
   self.closingTabIds.push(tab.id);
   if (closeParentBehavior == Constants.kCLOSE_PARENT_BEHAVIOR_CLOSE_ALL_CHILDREN) {
     self.closingDescendantTabIds = self.closingDescendantTabIds
-      .concat(Tree.getClosingTabsFromParent(tab).map(tab => tab.apiTab.id));
+      .concat(TreeBehavior.getClosingTabsFromParent(tab).map(tab => tab.id));
     self.closingDescendantTabIds = Array.from(new Set(self.closingDescendantTabIds));
   }
 
-  // this is required to wait until the closing tab is stored to the "recently closed" list
-  await wait(0);
   if (self.promisedGrantedToCloseTabs)
     return self.promisedGrantedToCloseTabs;
 
-  self.closingTabWasActive = self.closingTabWasActive || Tabs.isActive(tab);
+  self.closingTabWasActive = self.closingTabWasActive || tab.active;
 
   let shouldRestoreCount;
   self.promisedGrantedToCloseTabs = wait(10).then(async () => {
     self.closingTabIds = Array.from(new Set(self.closingTabIds));
     self.closingDescendantTabIds = self.closingDescendantTabIds.filter(id => !self.closingTabIds.includes(id))
     self.closingDescendantTabIds = Array.from(new Set(self.closingDescendantTabIds));
-    let allClosingTabs = [tab.apiTab.id].concat(self.closingDescendantTabIds);
+    let allClosingTabs = [tab].concat(self.closingDescendantTabIds.map(id => Tab.get(id)));
     allClosingTabs = Array.from(new Set(allClosingTabs));
     shouldRestoreCount = self.closingTabIds.length;
-    if (allClosingTabs.length > 0) {
+    const restorableClosingTabs = allClosingTabs.filter(tab => (
+      tab.url != 'about:blank' &&
+      tab.url != configs.guessNewOrphanTabAsOpenedByNewTabCommandUrl
+    ));
+    if (restorableClosingTabs.length > 0) {
       log('tryGrantClose: show confirmation for ', allClosingTabs);
-      return Background.confirmToCloseTabs(allClosingTabs, {
-        windowId: tab.apiTab.windowId
+      return Background.confirmToCloseTabs(allClosingTabs.map(tab => tab.$TST.sanitized), {
+        windowId: tab.windowId
       });
     }
     return true;
   })
     .then(async (granted) => {
       // remove the closed tab itself because it is already closed!
-      configs.grantedRemovingTabIds = configs.grantedRemovingTabIds.filter(id => id != tab.apiTab.id);
+      configs.grantedRemovingTabIds = configs.grantedRemovingTabIds.filter(id => id != tab.id);
       if (granted)
         return true;
       log(`tryGrantClose: not granted, restore ${shouldRestoreCount} tabs`);
-      const sessions = await browser.sessions.getRecentlyClosed({ maxResults: shouldRestoreCount * 2 });
-      const toBeRestoredTabs = [];
-      for (const session of sessions) {
-        if (!session.tab)
-          continue;
-        toBeRestoredTabs.push(session.tab);
-        if (toBeRestoredTabs.length == shouldRestoreCount)
-          break;
-      }
-      for (const tab of toBeRestoredTabs.reverse()) {
-        log('tryGrantClose: Tabrestoring session = ', tab);
-        browser.sessions.restore(tab.sessionId);
-        const tabs = await Tabs.waitUntilAllTabsAreCreated();
-        await Promise.all(tabs.map(tab => tab.opened));
-      }
+      // this is required to wait until the closing tab is stored to the "recently closed" list
+      wait(0).then(async () => {
+        const sessions = await browser.sessions.getRecentlyClosed({ maxResults: shouldRestoreCount * 2 }).catch(ApiTabs.  createErrorHandler());
+        const toBeRestoredTabs = [];
+        for (const session of sessions) {
+          if (!session.tab)
+            continue;
+          toBeRestoredTabs.push(session.tab);
+          if (toBeRestoredTabs.length == shouldRestoreCount)
+            break;
+        }
+        const promisedRestoredTabSets = [];
+        for (const tab of toBeRestoredTabs.reverse()) {
+          log('tryGrantClose: Tabrestoring session = ', tab);
+          promisedRestoredTabSets.push(Tab.doAndGetNewTabs(async () => {
+            browser.sessions.restore(tab.sessionId).catch(ApiTabs.createErrorSuppressor());
+            await Tab.waitUntilTrackedAll();
+          }));
+        }
+        const restoredTabs = (await Promise.all(promisedRestoredTabSets)).flat();
+        await Promise.all(restoredTabs.map(tab => tab && Tab.get(tab.id).$TST.opened));
+        log('tryGrantClose: restored ', restoredTabs);
+      });
       return false;
     });
 
@@ -174,7 +185,7 @@ tryGrantCloseTab.closingTabWasActive        = false;
 tryGrantCloseTab.promisedGrantedToCloseTabs = null;
 
 async function closeChildTabs(parent) {
-  const tabs = Tabs.getDescendantTabs(parent);
+  const tabs = parent.$TST.descendants;
   //if (!fireTabSubtreeClosingEvent(parent, tabs))
   //  return;
 
@@ -186,9 +197,9 @@ async function closeChildTabs(parent) {
   //fireTabSubtreeClosedEvent(parent, tabs);
 }
 
-Tabs.onRemoved.addListener((tab, info) => {
+Tab.onRemoved.addListener((tab, info) => {
   log('Tabs.onRemoved: removed ', dumpTab(tab));
-  configs.grantedRemovingTabIds = configs.grantedRemovingTabIds.filter(id => id != tab.apiTab.id);
+  configs.grantedRemovingTabIds = configs.grantedRemovingTabIds.filter(id => id != tab.id);
 
   if (info.isWindowClosing)
     return;
@@ -213,30 +224,21 @@ Tabs.onRemoved.addListener((tab, info) => {
 });
 
 browser.windows.onRemoved.addListener(windowId  => {
-  const removeTabIds = Tabs.getAllTabs(windowId).map(tab => tab.apiTab.id);
-  configs.grantedRemovingTabIds = configs.grantedRemovingTabIds.filter(id => !removeTabIds.includes(id));
+  const window = TabsStore.windows.get(windowId);
+  if (!window)
+    return;
+  configs.grantedRemovingTabIds = configs.grantedRemovingTabIds.filter(id => !window.tabs.has(id));
 });
 
 
-Tabs.onDetached.addListener((tab, info = {}) => {
-  if (typeof browser.tabs.moveInSuccession != 'function') { // on Firefox 64 or older
-    if (Tree.shouldApplyTreeBehavior(info)) {
-      Tree.tryMoveFocusFromClosingActiveTabNow(tab, {
-        ignoredTabs: Tabs.getDescendantTabs(tab)
-      });
-      return;
-    }
-
-    Tree.tryMoveFocusFromClosingActiveTab(tab);
-  }
-
+Tab.onDetached.addListener((tab, info = {}) => {
   log('Tabs.onDetached ', dumpTab(tab));
-  let closeParentBehavior = Tree.getCloseParentBehaviorForTabWithSidebarOpenState(tab, info);
+  let closeParentBehavior = TreeBehavior.getCloseParentBehaviorForTabWithSidebarOpenState(tab, info);
   if (closeParentBehavior == Constants.kCLOSE_PARENT_BEHAVIOR_CLOSE_ALL_CHILDREN)
     closeParentBehavior = Constants.kCLOSE_PARENT_BEHAVIOR_PROMOTE_FIRST_CHILD;
 
   Tree.detachAllChildren(tab, {
-    behavior: closeParentBehavior,
+    behavior:  closeParentBehavior,
     broadcast: true
   });
   //reserveCloseRelatedTabs(toBeClosedTabs);
