@@ -6,10 +6,14 @@
 import '../extlib/TabIdFixer.js';
 
 import {
-  log
+  log,
+  configs
 } from '/common/common.js';
 
 import * as MetricsData from '/common/metrics-data.js';
+import * as TabsStore from '/common/tabs-store.js';
+import * as SidebarConnection from '/common/sidebar-connection.js';
+import Tab from '/common/Tab.js';
 
 import * as Background from './background.js';
 import './handle-new-tabs.js';
@@ -40,3 +44,7 @@ window.dumpLogs = () => {
 // for old debugging method
 window.log = log;
 window.gMetricsData = MetricsData;
+window.Tab = Tab;
+window.TabsStore = TabsStore;
+window.SidebarConnection = SidebarConnection;
+window.configs = configs;
