@@ -7,12 +7,16 @@ import '/extlib/l10n.js';
 import '/extlib/TabIdFixer.js';
 
 import {
-  log
+  log,
+  configs
 } from '/common/common.js';
 
 import * as MetricsData from '/common/metrics-data.js';
+import * as TabsStore from '/common/tabs-store.js';
+import Tab from '/common/Tab.js';
 
 import * as Sidebar from './sidebar.js';
+import * as BackgroundConnection from './background-connection.js';
 import './mouse-event-listener.js';
 import './collapse-expand.js';
 
@@ -34,3 +38,7 @@ window.dumpLogs = () => {
 // for old debugging method
 window.log = log;
 window.gMetricsData = MetricsData;
+window.Tab = Tab;
+window.TabsStore = TabsStore;
+window.BackgroundConnection = BackgroundConnection;
+window.configs = configs;
