@@ -1,5 +1,5 @@
 /* ***** BEGIN LICENSE BLOCK ***** 
- * Version: MPL 1.1/GPL 2.0/LGPL 2.1
+ * Version: MPL 1.1
  *
  * The contents of this file are subject to the Mozilla Public License Version
  * 1.1 (the "License"); you may not use this file except in compliance with
@@ -14,7 +14,7 @@
  * The Original Code is the Tree Style Tab.
  *
  * The Initial Developer of the Original Code is YUKI "Piro" Hiroshi.
- * Portions created by the Initial Developer are Copyright (C) 2011-2017
+ * Portions created by the Initial Developer are Copyright (C) 2011-2019
  * the Initial Developer. All Rights Reserved.
  *
  * Contributor(s): YUKI "Piro" Hiroshi <piro.outsider.reflex@gmail.com>
@@ -23,18 +23,6 @@
  *                 Xidorn Quan <https://github.com/upsuper> (Firefox 40+ support)
  *                 lv7777 (https://github.com/lv7777)
  *
- * Alternatively, the contents of this file may be used under the terms of
- * either the GNU General Public License Version 2 or later (the "GPL"), or
- * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
- * in which case the provisions of the GPL or the LGPL are applicable instead
- * of those above. If you wish to allow use of your version of this file only
- * under the terms of either the GPL or the LGPL, and not to allow others to
- * use your version of this file under the terms of the MPL, indicate your
- * decision by deleting the provisions above and replace them with the notice
- * and other provisions required by the GPL or the LGPL. If you do not delete
- * the provisions above, a recipient may use your version of this file under
- * the terms of any one of the MPL, the GPL or the LGPL.
- *
  * ***** END LICENSE BLOCK ******/
 'use strict';
 
@@ -43,16 +31,20 @@ import MenuUI from '/extlib/MenuUI.js';
 import {
   log as internalLogger,
   wait,
+  dumpTab,
   configs
 } from '/common/common.js';
 import * as Constants from '/common/constants.js';
-import * as Tabs from '/common/tabs.js';
+import * as ApiTabs from '/common/api-tabs.js';
+import * as TabsStore from '/common/tabs-store.js';
 import * as TabsInternalOperation from '/common/tabs-internal-operation.js';
-import * as Tree from '/common/tree.js';
+import * as TreeBehavior from '/common/tree-behavior.js';
 import * as TSTAPI from '/common/tst-api.js';
-import * as Commands from '/common/commands.js';
 import * as MetricsData from '/common/metrics-data.js';
 
+import Tab from '/common/Tab.js';
+
+import * as BackgroundConnection from './background-connection.js';
 import * as Sidebar from './sidebar.js';
 import * as EventUtils from './event-utils.js';
 import * as DragAndDrop from './drag-and-drop.js';
@@ -68,11 +60,8 @@ const mTabBar = document.querySelector('#tabbar');
 const mContextualIdentitySelector = document.getElementById(Constants.kCONTEXTUAL_IDENTITY_SELECTOR);
 const mNewTabActionSelector       = document.getElementById(Constants.kNEWTAB_ACTION_SELECTOR);
 
-const mUpdatingCollapsedStateCancellers = new WeakMap();
-const mTabCollapsedStateChangedManagers = new WeakMap();
-
 Sidebar.onInit.addListener(() => {
-  mTargetWindow = Tabs.getWindow();
+  mTargetWindow = TabsStore.getWindow();
 });
 
 Sidebar.onBuilt.addListener(async () => {
@@ -81,9 +70,10 @@ Sidebar.onBuilt.addListener(async () => {
   document.addEventListener('click', onClick);
   mTabBar.addEventListener('dblclick', onDblClick);
 
-  MetricsData.add('apply configs');
+  MetricsData.add('mouse-event-listener: Sidebar.onBuilt: apply configs');
 
   browser.runtime.onMessage.addListener(onMessage);
+  BackgroundConnection.onMessage.addListener(onBackgroundMessage);
 
   mContextualIdentitySelector.ui = new MenuUI({
     root:       mContextualIdentitySelector,
@@ -150,7 +140,7 @@ function onMouseMove(event) {
   if (tab) {
     TSTAPI.sendMessage({
       type:     TSTAPI.kNOTIFY_TAB_MOUSEMOVE,
-      tab:      TSTAPI.serializeTab(tab),
+      tab:      new TSTAPI.TreeItem(tab),
       window:   mTargetWindow,
       windowId: mTargetWindow,
       ctrlKey:  event.ctrlKey,
@@ -158,7 +148,7 @@ function onMouseMove(event) {
       altKey:   event.altKey,
       metaKey:  event.metaKey,
       dragging: DragAndDrop.isCapturingForDragging()
-    });
+    }, { tabProperties: ['tab'] }).catch(_error => {});
   }
 }
 onMouseMove = EventUtils.wrapWithErrorHandler(onMouseMove);
@@ -171,12 +161,12 @@ function onMouseOver(event) {
   // contains the target of the mouseout event or null if there is none). This
   // also includes the case where we enter the tab directly without going
   // through another tab or the sidebar, which causes relatedTarget to be null
-  const enterTabFromAncestor = tab && !tab.contains(event.relatedTarget);
+  const enterTabFromAncestor = tab && !tab.$TST.element.contains(event.relatedTarget);
 
   if (enterTabFromAncestor) {
     TSTAPI.sendMessage({
       type:     TSTAPI.kNOTIFY_TAB_MOUSEOVER,
-      tab:      TSTAPI.serializeTab(tab),
+      tab:      new TSTAPI.TreeItem(tab),
       window:   mTargetWindow,
       windowId: mTargetWindow,
       ctrlKey:  event.ctrlKey,
@@ -184,7 +174,7 @@ function onMouseOver(event) {
       altKey:   event.altKey,
       metaKey:  event.metaKey,
       dragging: DragAndDrop.isCapturingForDragging()
-    });
+    }, { tabProperties: ['tab'] }).catch(_error => {});
   }
 }
 onMouseOver = EventUtils.wrapWithErrorHandler(onMouseOver);
@@ -197,12 +187,12 @@ function onMouseOut(event) {
   // contains the target of the mouseover event or null if there is none). This
   // also includes the case where we leave the tab directly without going
   // through another tab or the sidebar, which causes relatedTarget to be null
-  const leaveTabToAncestor = tab && !tab.contains(event.relatedTarget);
+  const leaveTabToAncestor = tab && !tab.$TST.element.contains(event.relatedTarget);
 
   if (leaveTabToAncestor) {
     TSTAPI.sendMessage({
       type:     TSTAPI.kNOTIFY_TAB_MOUSEOUT,
-      tab:      TSTAPI.serializeTab(tab),
+      tab:      new TSTAPI.TreeItem(tab),
       window:   mTargetWindow,
       windowId: mTargetWindow,
       ctrlKey:  event.ctrlKey,
@@ -210,7 +200,7 @@ function onMouseOut(event) {
       altKey:   event.altKey,
       metaKey:  event.metaKey,
       dragging: DragAndDrop.isCapturingForDragging()
-    });
+    }, { tabProperties: ['tab'] }).catch(_error => {});
   }
 }
 onMouseOut = EventUtils.wrapWithErrorHandler(onMouseOut);
@@ -241,6 +231,9 @@ function onMouseDown(event) {
   const mousedownDetail = {
     targetType:    getMouseEventTargetType(event),
     tab:           tab && tab.id,
+    tabId:         tab && tab.id,
+    window:        mTargetWindow,
+    windowId:      mTargetWindow,
     twisty:        EventUtils.isEventFiredOnTwisty(event),
     soundButton:   EventUtils.isEventFiredOnSoundButton(event),
     closebox:      EventUtils.isEventFiredOnClosebox(event),
@@ -265,14 +258,23 @@ function onMouseDown(event) {
 
   const mousedown = {
     detail: mousedownDetail,
+    treeItem: new TSTAPI.TreeItem(tab),
     promisedMousedownNotified: Promise.resolve()
   };
 
-  mousedown.promisedMousedownNotified = browser.runtime.sendMessage(Object.assign({}, mousedownDetail, {
-    type:     Constants.kNOTIFY_TAB_MOUSEDOWN,
-    window:   mTargetWindow,
-    windowId: mTargetWindow,
-  }));
+  mousedown.promisedMousedownNotified = Promise.all([
+    browser.runtime.sendMessage({type: Constants.kNOTIFY_TAB_MOUSEDOWN })
+      .catch(ApiTabs.createErrorHandler()),
+    (async () => {
+      if (mousedownDetail.targetType != 'tab')
+        return;
+      log('Sending message to listeners');
+      return TSTAPI.sendMessage(Object.assign({}, mousedownDetail, {
+        type: TSTAPI.kNOTIFY_TAB_MOUSEDOWN,
+        tab:  mousedown.treeItem
+      }), { tabProperties: ['tab'] });
+    })()
+  ]).then(results => results[1]);
 
   EventUtils.setLastMousedown(event.button, mousedown);
   mousedown.timeout = setTimeout(async () => {
@@ -298,16 +300,8 @@ function onMouseDown(event) {
 
     if (event.button == 0 &&
         tab) {
-      const results = await DragAndDrop.legacyStartMultiDrag(tab, mousedown.detail.closebox);
-      if (results.some(result => result.result !== false)) {
-        log('onMouseDown expired');
-        mousedown.expired = true;
-        onMessage({
-          type:     Constants.kNOTIFY_TAB_MOUSEDOWN_EXPIRED,
-          windowId: mTargetWindow,
-          button:   event.button
-        });
-      }
+      log('onMouseDown expired');
+      mousedown.expired = true;
     }
   }, configs.longPressDuration);
 }
@@ -338,11 +332,11 @@ function getMouseEventTargetType(event) {
 }
 
 async function onMouseUp(event) {
-  const tab = EventUtils.getTabFromEvent(event, { force: true }) || EventUtils.getTabFromTabbarEvent(event, { force: true });
-  const livingTab = EventUtils.getTabFromEvent(event);
-  log('onMouseUp: ', tab, { living: !!livingTab });
+  const unsafeTab = EventUtils.getTabFromEvent(event, { force: true }) || EventUtils.getTabFromTabbarEvent(event, { force: true });
+  const tab       = EventUtils.getTabFromEvent(event) || EventUtils.getTabFromTabbarEvent(event);
+  log('onMouseUp: ', unsafeTab, { living: !!tab });
 
-  DragAndDrop.endMultiDrag(livingTab, event);
+  DragAndDrop.endMultiDrag(unsafeTab, event);
 
   if (EventUtils.isEventFiredOnMenuOrPanel(event) ||
       EventUtils.isEventFiredOnAnchor(event))
@@ -350,101 +344,48 @@ async function onMouseUp(event) {
 
   const lastMousedown = EventUtils.getLastMousedown(event.button);
   EventUtils.cancelHandleMousedown(event.button);
-  if (lastMousedown)
-    await lastMousedown.promisedMousedownNotified;
-
-  const serializedTab = livingTab && TSTAPI.serializeTab(livingTab);
-  let promisedCanceled = Promise.resolve(false);
-  if (serializedTab && lastMousedown) {
-    const results = TSTAPI.sendMessage(Object.assign({}, lastMousedown.detail, {
-      type:    TSTAPI.kNOTIFY_TAB_MOUSEUP,
-      tab:     serializedTab,
-      window:  mTargetWindow,
-      windowId: mTargetWindow
-    }));
-    // don't wait here, because we need process following common operations
-    // even if this mouseup event is canceled.
-    promisedCanceled = results.then(results => results.some(result => result.result));
-  }
-
-  if (!lastMousedown ||
-      lastMousedown.expired ||
-      lastMousedown.detail.targetType != getMouseEventTargetType(event) ||
-      (livingTab && livingTab != Tabs.getTabById(lastMousedown.detail.tab)))
+  if (!lastMousedown)
     return;
 
-  log('onMouseUp ', lastMousedown.detail);
+  let promisedCanceled = null;
+  if (lastMousedown.treeItem && lastMousedown.detail.targetType == 'tab')
+    promisedCanceled = Promise.all([
+      TSTAPI.sendMessage(Object.assign({}, lastMousedown.detail, {
+        type: TSTAPI.kNOTIFY_TAB_MOUSEUP,
+        tab:  lastMousedown.treeItem
+      }), { tabProperties: ['tab'] }),
 
-  if (await promisedCanceled) {
+      TSTAPI.sendMessage(Object.assign({}, lastMousedown.detail, {
+        type: TSTAPI.kNOTIFY_TAB_CLICKED,
+        tab:  lastMousedown.treeItem
+      }), { tabProperties: ['tab'] }),
+
+      lastMousedown.promisedMousedownNotified
+    ])
+      .then(results => results.flat())
+      .then(results => results.some(result => result && result.result));
+
+  if (lastMousedown.expired ||
+      lastMousedown.detail.targetType != getMouseEventTargetType(event) || // when the cursor was moved before mouseup
+      (tab && tab != Tab.get(lastMousedown.detail.tab))) // when the tab was already removed
+    return;
+
+  if (promisedCanceled && await promisedCanceled) {
     log('onMouseUp: canceled / by other addons');
     return;
   }
 
-  if (livingTab) {
-    if (lastMousedown.detail.isMiddleClick) { // Ctrl-click doesn't close tab on Firefox's tab bar!
-      log('onMouseUp: middle click on a tab');
-      //log('middle-click to close');
-      const tabs = Tree.getClosingTabsFromParent(livingTab);
-      Sidebar.confirmToCloseTabs(tabs.map(tab => tab.apiTab.id))
-        .then(confirmed => {
-          if (confirmed)
-            TabsInternalOperation.removeTab(livingTab, { inRemote: true });
-        });
-    }
-    else if (lastMousedown.detail.twisty) {
-      log('clicked on twisty');
-      if (Tabs.hasChildTabs(tab))
-        Tree.collapseExpandSubtree(tab, {
-          collapsed:       !Tabs.isSubtreeCollapsed(tab),
-          manualOperation: true,
-          inRemote:        true
-        });
-    }
-    else if (lastMousedown.detail.soundButton) {
-      log('clicked on sound button');
-      browser.runtime.sendMessage({
-        type:     Constants.kCOMMAND_SET_SUBTREE_MUTED,
-        windowId: mTargetWindow,
-        tab:      tab.id,
-        muted:    Tabs.maybeSoundPlaying(tab)
-      });
-    }
-    else if (lastMousedown.detail.closebox) {
-      log('clicked on closebox');
-      //if (!warnAboutClosingTabSubtreeOf(tab)) {
-      //  event.stopPropagation();
-      //  event.preventDefault();
-      //  return;
-      //}
-      const multiselected  = Tabs.isMultiselected(tab);
-      const tabsToBeClosed = multiselected ?
-        Tabs.getSelectedTabs(tab) :
-        Tree.getClosingTabsFromParent(tab) ;
-      Sidebar.confirmToCloseTabs(tabsToBeClosed.map(tab => tab.apiTab.id))
-        .then(confirmed => {
-          if (!confirmed)
-            return;
-          if (multiselected)
-            TabsInternalOperation.removeTabs(tabsToBeClosed, { inRemote: true });
-          else
-            TabsInternalOperation.removeTab(tab, { inRemote: true });
-        });
-    }
-    else if (lastMousedown.detail.button == 0 &&
-             !lastMousedown.detail.altKey &&
-             !lastMousedown.detail.ctrlKey &&
-             !lastMousedown.detail.metaKey &&
-             !lastMousedown.detail.shiftKey) {
-      // clear selection by left click
-      browser.tabs.highlight({
-        windowId: tab.apiTab.windowId,
-        tabs:     [tab.apiTab.index],
-        populate: false
-      }).catch(_e => {});
-    }
-    return;
-  }
-  else if (tab) // ignore mouseup on closing tab or something
+  // not canceled, then fallback to default behavior
+  return handleDefaultMouseUp({ lastMousedown, tab, event });
+}
+onMouseUp = EventUtils.wrapWithErrorHandler(onMouseUp);
+
+async function handleDefaultMouseUp({ lastMousedown, tab, event }) {
+  log('handleDefaultMouseUp ', lastMousedown.detail);
+
+  if (tab &&
+      lastMousedown.detail.button != 2 &&
+      await handleDefaultMouseUpOnTab(lastMousedown, tab))
     return;
 
   // following codes are for handlig of click event on the tab bar itself.
@@ -463,15 +404,17 @@ async function onMouseUp(event) {
 
   log('onMouseUp: notify as a blank area click to other addons');
   let results = await TSTAPI.sendMessage(Object.assign({}, lastMousedown.detail, {
-    type:   TSTAPI.kNOTIFY_TABBAR_MOUSEUP,
-    window: mTargetWindow,
-    windowId: mTargetWindow
-  }));
+    type:     TSTAPI.kNOTIFY_TABBAR_MOUSEUP,
+    window:   mTargetWindow,
+    windowId: mTargetWindow,
+    tab:      lastMousedown.treeItem
+  }), { tabProperties: ['tab'] });
   results = results.concat(await TSTAPI.sendMessage(Object.assign({}, lastMousedown.detail, {
-    type:   TSTAPI.kNOTIFY_TABBAR_CLICKED,
-    window: mTargetWindow,
-    windowId: mTargetWindow
-  })));
+    type:     TSTAPI.kNOTIFY_TABBAR_CLICKED,
+    window:   mTargetWindow,
+    windowId: mTargetWindow,
+    tab:      lastMousedown.treeItem
+  }), { tabProperties: ['tab'] }));
   if (results.some(result => result.result))// canceled
     return;
 
@@ -482,7 +425,234 @@ async function onMouseUp(event) {
     });
   }
 }
-onMouseUp = EventUtils.wrapWithErrorHandler(onMouseUp);
+handleDefaultMouseUp = EventUtils.wrapWithErrorHandler(handleDefaultMouseUp);
+
+async function handleDefaultMouseUpOnTab(lastMousedown, tab) {
+  log('Ready to handle click action on the tab');
+
+  const onRegularArea = (
+    !lastMousedown.detail.twisty &&
+    !lastMousedown.detail.soundButton &&
+    !lastMousedown.detail.closebox
+  );
+  const wasMultiselectionAction = (
+    onRegularArea &&
+    updateMultiselectionByTabClick(tab, lastMousedown.detail)
+  );
+  log(' => ', { onRegularArea, wasMultiselectionAction });
+
+  if (lastMousedown.detail.button == 0 &&
+      onRegularArea &&
+      !wasMultiselectionAction)
+    TabsInternalOperation.activateTab(tab, {
+      keepMultiselection: false // tab.highlighted
+    });
+
+  if (lastMousedown.detail.isMiddleClick) { // Ctrl-click doesn't close tab on Firefox's tab bar!
+    log('onMouseUp: middle click on a tab');
+    if (lastMousedown.detail.targetType != 'tab') // ignore middle click on blank area
+      return false;
+    const tabs = TreeBehavior.getClosingTabsFromParent(tab, {
+      byInternalOperation: true
+    });
+    Sidebar.confirmToCloseTabs(tabs.map(tab => tab.$TST.sanitized))
+      .then(confirmed => {
+        if (confirmed)
+          BackgroundConnection.sendMessage({
+            type:   Constants.kCOMMAND_REMOVE_TABS_INTERNALLY,
+            tabIds: [tab.id]
+          });
+      });
+  }
+  else if (lastMousedown.detail.twisty) {
+    log('clicked on twisty');
+    if (tab.$TST.hasChild)
+      BackgroundConnection.sendMessage({
+        type:            Constants.kCOMMAND_SET_SUBTREE_COLLAPSED_STATE,
+        tabId:           tab.id,
+        collapsed:       !tab.$TST.subtreeCollapsed,
+        manualOperation: true,
+        stack:           configs.debug && new Error().stack
+      });
+  }
+  else if (lastMousedown.detail.soundButton) {
+    log('clicked on sound button');
+    BackgroundConnection.sendMessage({
+      type:  Constants.kCOMMAND_SET_SUBTREE_MUTED,
+      tabId: tab.id,
+      muted: tab.$TST.maybeSoundPlaying
+    });
+  }
+  else if (lastMousedown.detail.closebox) {
+    log('clicked on closebox');
+    //if (!warnAboutClosingTabSubtreeOf(tab)) {
+    //  event.stopPropagation();
+    //  event.preventDefault();
+    //  return;
+    //}
+    const multiselected  = tab.$TST.multiselected;
+    const tabsToBeClosed = multiselected ?
+      Tab.getSelectedTabs(tab.windowId) :
+      TreeBehavior.getClosingTabsFromParent(tab, {
+        byInternalOperation: true
+      }) ;
+    Sidebar.confirmToCloseTabs(tabsToBeClosed.map(tab => tab.$TST.sanitized))
+      .then(confirmed => {
+        if (!confirmed)
+          return;
+        BackgroundConnection.sendMessage({
+          type:   Constants.kCOMMAND_REMOVE_TABS_INTERNALLY,
+          tabIds: multiselected ? tabsToBeClosed.map(tab => tab.id) : [tab.id]
+        });
+      });
+  }
+
+  return true;
+}
+handleDefaultMouseUpOnTab = EventUtils.wrapWithErrorHandler(handleDefaultMouseUpOnTab);
+
+const mLastClickedTabInWindow = new Map();
+const mIsInSelectionSession   = new Map();
+
+function updateMultiselectionByTabClick(tab, event) {
+  const ctrlKeyPressed     = event.ctrlKey || (event.metaKey && /^Mac/i.test(navigator.platform));
+  const activeTab          = Tab.getActiveTab(tab.windowId);
+  const highlightedTabIds  = new Set(Tab.getHighlightedTabs(tab.windowId).map(tab => tab.id));
+  const inSelectionSession = mIsInSelectionSession.get(tab.windowId);
+  log('updateMultiselectionByTabClick ', { ctrlKeyPressed, activeTab, highlightedTabIds, inSelectionSession });
+  if (event.shiftKey) {
+    // select the clicked tab and tabs between last activated tab
+    const lastClickedTab   = mLastClickedTabInWindow.get(tab.windowId) || activeTab;
+    const betweenTabs      = Tab.getTabsBetween(lastClickedTab, tab);
+    const targetTabs       = new Set([lastClickedTab].concat(betweenTabs));
+    targetTabs.add(tab);
+
+    log(' => ', { lastClickedTab, betweenTabs, targetTabs });
+
+    try {
+      if (!ctrlKeyPressed) {
+        const alreadySelectedTabs = Tab.getSelectedTabs(tab.windowId, { iterator: true });
+        log('clear old selection by shift-click: ', configs.debug && Array.from(alreadySelectedTabs, dumpTab));
+        for (const alreadySelectedTab of alreadySelectedTabs) {
+          if (!targetTabs.has(alreadySelectedTab))
+            highlightedTabIds.delete(alreadySelectedTab.id);
+        }
+      }
+
+      log('set selection by shift-click: ', configs.debug && Array.from(targetTabs, dumpTab));
+      for (const toBeSelectedTab of targetTabs) {
+        highlightedTabIds.add(toBeSelectedTab.id);
+      }
+
+      const rootTabs = [tab];
+      if (tab != activeTab &&
+          !inSelectionSession)
+        rootTabs.push(activeTab);
+      for (const root of rootTabs) {
+        if (!root.$TST.subtreeCollapsed)
+          continue;
+        for (const descendant of root.$TST.descendants) {
+          highlightedTabIds.add(descendant.id);
+        }
+      }
+
+      // for better performance, we should not call browser.tabs.update() for each tab.
+      const indices = Array.from(highlightedTabIds)
+        .filter(id => id != activeTab.id)
+        .map(id => Tab.get(id).index);
+      if (highlightedTabIds.has(activeTab.id))
+        indices.unshift(activeTab.index);
+      browser.tabs.highlight({
+        windowId: tab.windowId,
+        populate: false,
+        tabs:     indices
+      }).catch(ApiTabs.createErrorSuppressor());
+    }
+    catch(_e) { // not implemented on old Firefox
+      return false;
+    }
+    mIsInSelectionSession.set(tab.windowId, true);
+    return true;
+  }
+  else if (ctrlKeyPressed) {
+    try {
+      log('change selection by ctrl-click: ', dumpTab(tab));
+      /* Special operation to toggle selection of collapsed descendants for the active tab.
+         - When there is no other multiselected foreign tab
+           => toggle multiselection only descendants.
+         - When there is one or more multiselected foreign tab
+           => toggle multiselection of the active tab and descendants.
+              => one of multiselected foreign tabs will be activated.
+         - When a foreign tab is highlighted and there is one or more unhighlighted descendants 
+           => highlight all descendants (to prevent only the root tab is dragged).
+       */
+      const activeTabDescendants = activeTab.$TST.descendants;
+      let toBeHighlighted = !tab.highlighted;
+      log('toBeHighlighted: ', toBeHighlighted);
+      if (tab == activeTab &&
+          tab.$TST.subtreeCollapsed &&
+          activeTabDescendants.length > 0) {
+        const highlightedCount  = activeTabDescendants.filter(tab => tab.highlighted).length;
+        const partiallySelected = highlightedCount != 0 && highlightedCount != activeTabDescendants.length;
+        toBeHighlighted = partiallySelected || !activeTabDescendants[0].highlighted;
+        log(' => ', toBeHighlighted, { partiallySelected });
+      }
+      if (toBeHighlighted)
+        highlightedTabIds.add(tab.id);
+      else
+        highlightedTabIds.delete(tab.id);
+
+      if (tab.$TST.subtreeCollapsed) {
+        const descendants = tab == activeTab ? activeTabDescendants : tab.$TST.descendants;
+        for (const descendant of descendants) {
+          if (toBeHighlighted)
+            highlightedTabIds.add(descendant.id);
+          else
+            highlightedTabIds.delete(descendant.id);
+        }
+      }
+
+      if (tab == activeTab) {
+        if (highlightedTabIds.size == 0) {
+          log('Don\'t unhighlight only one highlighted active tab!');
+          highlightedTabIds.add(tab.id);
+        }
+      }
+      else if (!inSelectionSession) {
+        log('Select active tab and its descendants, for new selection session');
+        highlightedTabIds.add(activeTab.id);
+        if (activeTab.$TST.subtreeCollapsed) {
+          for (const descendant of activeTabDescendants) {
+            highlightedTabIds.add(descendant.id);
+          }
+        }
+      }
+
+      // for better performance, we should not call browser.tabs.update() for each tab.
+      const indices = Array.from(highlightedTabIds)
+        .filter(id => id != activeTab.id)
+        .map(id => Tab.get(id).index);
+      if (highlightedTabIds.has(activeTab.id))
+        indices.unshift(activeTab.index);
+      browser.tabs.highlight({
+        windowId: tab.windowId,
+        populate: false,
+        tabs:     indices
+      }).catch(ApiTabs.createErrorSuppressor());
+    }
+    catch(_e) { // not implemented on old Firefox
+      return false;
+    }
+    mLastClickedTabInWindow.set(tab.windowId, tab);
+    mIsInSelectionSession.set(tab.windowId, true);
+    return true;
+  }
+  else {
+    mLastClickedTabInWindow.set(tab.windowId, tab);
+    mIsInSelectionSession.delete(tab.windowId);
+    return false;
+  }
+}
 
 function onClick(_event) {
   // clear unexpectedly left "dragging" state
@@ -497,12 +667,12 @@ function handleNewTabAction(event, options = {}) {
   if (!configs.autoAttach && !('action' in options))
     options.action = Constants.kNEWTAB_DO_NOTHING;
 
-  Commands.openNewTabAs({
-    baseTab:      Tabs.getActiveTab(mTargetWindow),
-    as:           options.action,
+  BackgroundConnection.sendMessage({
+    type:          Constants.kCOMMAND_NEW_TAB_AS,
+    baseTabId:     TabsStore.activeTabInWindow.get(mTargetWindow).id,
+    as:            options.action,
     cookieStoreId: options.cookieStoreId,
-    inBackground: event.shiftKey,
-    inRemote:     true
+    inBackground:  event.shiftKey
   });
 }
 
@@ -523,15 +693,20 @@ function onDblClick(event) {
         !event.shiftKey) {
       event.stopPropagation();
       event.preventDefault();
-      TabsInternalOperation.removeTab(livingTab, { inRemote: true });
+      BackgroundConnection.sendMessage({
+        type:   Constants.kCOMMAND_REMOVE_TABS_INTERNALLY,
+        tabIds: [livingTab.id],
+      });
     }
     else if (configs.collapseExpandSubtreeByDblClick) {
       event.stopPropagation();
       event.preventDefault();
-      Tree.collapseExpandSubtree(livingTab, {
-        collapsed:       !Tabs.isSubtreeCollapsed(livingTab),
+      BackgroundConnection.sendMessage({
+        type:            Constants.kCOMMAND_SET_SUBTREE_COLLAPSED_STATE,
+        tabId:           livingTab.id,
+        collapsed:       !livingTab.$TST.subtreeCollapsed,
         manualOperation: true,
-        inRemote:        true
+        stack:           configs.debug && new Error().stack
       });
     }
     return;
@@ -583,13 +758,6 @@ function onContextualIdentitySelect(item, event) {
 }
 
 
-
-Tabs.onRemoved.addListener(async (tab, _info) => {
-  mUpdatingCollapsedStateCancellers.delete(tab);
-  mTabCollapsedStateChangedManagers.delete(tab);
-});
-
-
 function onMessage(message, _sender, _respond) {
   if (!message ||
       typeof message.type != 'string' ||
@@ -598,11 +766,22 @@ function onMessage(message, _sender, _respond) {
 
   //log('onMessage: ', message, sender);
   switch (message.type) {
-    case Constants.kNOTIFY_TAB_MOUSEDOWN_CANCELED:
-      if (message.windowId == mTargetWindow)
-        EventUtils.cancelHandleMousedown(message.button || 0);
+    case TSTAPI.kCOMMAND_BROADCAST_API_REGISTERED:
+      wait(0).then(() => { // wait until addons are updated
+        updateSpecialEventListenersForAPIListeners();
+      });
       break;
 
+    case TSTAPI.kCOMMAND_BROADCAST_API_UNREGISTERED:
+      wait(0).then(() => { // wait until addons are updated
+        updateSpecialEventListenersForAPIListeners();
+      });
+      break;
+  }
+}
+
+function onBackgroundMessage(message) {
+  switch (message.type) {
     case Constants.kNOTIFY_TAB_MOUSEDOWN_EXPIRED:
       if (message.windowId == mTargetWindow) {
         const lastMousedown = EventUtils.getLastMousedown(message.button || 0);
@@ -618,17 +797,5 @@ function onMessage(message, _sender, _respond) {
       `);
       mContextualIdentitySelector.ui.open({ anchor });
     }; break;
-
-    case TSTAPI.kCOMMAND_BROADCAST_API_REGISTERED:
-      wait(0).then(() => { // wait until addons are updated
-        updateSpecialEventListenersForAPIListeners();
-      });
-      break;
-
-    case TSTAPI.kCOMMAND_BROADCAST_API_UNREGISTERED:
-      wait(0).then(() => { // wait until addons are updated
-        updateSpecialEventListenersForAPIListeners();
-      });
-      break;
   }
 }
