@@ -5,12 +5,13 @@
 */
 'use strict';
 
-import * as Constants from './constants.js';
 import {
   log as internalLogger,
   notify,
   configs
 } from './common.js';
+import * as Constants from './constants.js';
+import * as ApiTabs from '/common/api-tabs.js';
 
 function log(...args) {
   internalLogger('common/permissions', ...args);
@@ -18,8 +19,7 @@ function log(...args) {
 
 export const BOOKMARKS = { permissions: ['bookmarks'] };
 export const ALL_URLS = { origins: ['<all_urls>'] };
-export const SEARCH = { permissions: ['search'] };
-export const MENUS_OVERRIDE_CONTEXT = { permissions: ['menus.overrideContext'] };
+export const TAB_HIDE = { permissions: ['tabHide'] };
 
 export function clearRequest() {
   configs.requestingPermissions = null;
@@ -27,7 +27,7 @@ export function clearRequest() {
 
 export function isGranted(permissions) {
   try {
-    return browser.permissions.contains(permissions);
+    return browser.permissions.contains(permissions).catch(ApiTabs.createErrorHandler());
   }
   catch(_e) {
     return Promise.reject(new Error('unsupported permission'));
@@ -77,7 +77,7 @@ export function bindToCheckbox(permissions, checkbox, options = {}) {
   checkbox.requestPermissions = async () => {
     try {
       if (!checkbox.checked) {
-        await browser.permissions.remove(permissions);
+        await browser.permissions.remove(permissions).catch(ApiTabs.createErrorSuppressor());
         if (options.onChanged)
           options.onChanged(false);
         return;
@@ -87,19 +87,9 @@ export function bindToCheckbox(permissions, checkbox, options = {}) {
       if (configs.requestingPermissionsNatively)
         return;
 
-      // Following code will throw error on Firefox 60 and earlier (but not on Firefox ESR 60)
-      // due to https://bugzilla.mozilla.org/show_bug.cgi?id=1382953
-      // Also must not have used await before calling browser.permissions.request or it will throw an error.
-      let granted;
-      try {
-        configs.requestingPermissionsNatively = permissions;
-        granted = await browser.permissions.request(permissions);
-      }
-      catch (_error) {
-      }
-      finally {
-        configs.requestingPermissionsNatively = null;
-      }
+      configs.requestingPermissionsNatively = permissions;
+      let granted = await browser.permissions.request(permissions).catch(ApiTabs.createErrorHandler());
+      configs.requestingPermissionsNatively = null;
 
       if (granted === undefined)
         granted = await isGranted(permissions);
@@ -147,8 +137,9 @@ export function requestPostProcess() {
         browser.runtime.sendMessage({
           type:        Constants.kCOMMAND_NOTIFY_PERMISSIONS_GRANTED,
           permissions: permissions
-        });
+        }).catch(_error => {});
     })
+    .catch(ApiTabs.createErrorSuppressor())
     .finally(() => {
       configs.requestingPermissionsNatively = null;
     });
