@@ -1,5 +1,5 @@
 /* ***** BEGIN LICENSE BLOCK ***** 
- * Version: MPL 1.1/GPL 2.0/LGPL 2.1
+ * Version: MPL 1.1
  *
  * The contents of this file are subject to the Mozilla Public License Version
  * 1.1 (the "License"); you may not use this file except in compliance with
@@ -14,7 +14,7 @@
  * The Original Code is the Tree Style Tab.
  *
  * The Initial Developer of the Original Code is YUKI "Piro" Hiroshi.
- * Portions created by the Initial Developer are Copyright (C) 2011-2017
+ * Portions created by the Initial Developer are Copyright (C) 2011-2019
  * the Initial Developer. All Rights Reserved.
  *
  * Contributor(s): YUKI "Piro" Hiroshi <piro.outsider.reflex@gmail.com>
@@ -23,18 +23,6 @@
  *                 Xidorn Quan <https://github.com/upsuper> (Firefox 40+ support)
  *                 lv7777 (https://github.com/lv7777)
  *
- * Alternatively, the contents of this file may be used under the terms of
- * either the GNU General Public License Version 2 or later (the "GPL"), or
- * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
- * in which case the provisions of the GPL or the LGPL are applicable instead
- * of those above. If you wish to allow use of your version of this file only
- * under the terms of either the GPL or the LGPL, and not to allow others to
- * use your version of this file under the terms of the MPL, indicate your
- * decision by deleting the provisions above and replace them with the notice
- * and other provisions required by the GPL or the LGPL. If you do not delete
- * the provisions above, a recipient may use your version of this file under
- * the terms of any one of the MPL, the GPL or the LGPL.
- *
  * ***** END LICENSE BLOCK ******/
 'use strict';
 
@@ -42,17 +30,21 @@ import {
   log as internalLogger,
   wait,
   nextFrame,
-  dumpTab,
   configs
 } from '/common/common.js';
 
 import * as Constants from '/common/constants.js';
-import * as Tabs from '/common/tabs.js';
-import * as Tree from '/common/tree.js';
+import * as ApiTabs from '/common/api-tabs.js';
+import * as TabsStore from '/common/tabs-store.js';
 import * as TSTAPI from '/common/tst-api.js';
 
+import Tab from '/common/Tab.js';
+
 import * as Size from './size.js';
 import * as EventUtils from './event-utils.js';
+import * as BackgroundConnection from './background-connection.js';
+import * as SidebarTabs from './sidebar-tabs.js';
+import * as CollapseExpand from './collapse-expand.js';
 
 import * as RestoringTabCount from './restoring-tab-count.js';
 
@@ -64,11 +56,9 @@ function log(...args) {
 let mTabBar;
 let mOutOfViewTabNotifier;
 
-export async function init() {
+export function init(scrollPosition) {
   mTabBar               = document.querySelector('#tabbar');
   mOutOfViewTabNotifier = document.querySelector('#out-of-view-tab-notifier');
-
-  const scrollPosition = await browser.sessions.getWindowValue(Tabs.getWindow(), Constants.kWINDOW_STATE_SCROLL_POSITION);
   if (typeof scrollPosition == 'number') {
     log('restore scroll position');
     cancelRunningScroll();
@@ -80,7 +70,7 @@ export async function init() {
 
   document.addEventListener('wheel', onWheel, { capture: true });
   mTabBar.addEventListener('scroll', onScroll);
-  browser.runtime.onMessage.addListener(onMessage);
+  BackgroundConnection.onMessage.addListener(onBackgroundMessage);
   browser.runtime.onMessageExternal.addListener(onMessageExternal);
 }
 
@@ -109,10 +99,11 @@ function cancelRunningScroll() {
 }
 
 function calculateScrollDeltaForTab(tab) {
-  if (Tabs.isPinned(tab))
+  tab = Tab.get(tab && tab.id);
+  if (!tab || tab.pinned)
     return 0;
 
-  const tabRect       = tab.getBoundingClientRect();
+  const tabRect       = tab.$TST.element.getBoundingClientRect();
   const containerRect = mTabBar.getBoundingClientRect();
   const offset        = getOffsetForAnimatingTab(tab) + smoothScrollTo.currentOffset;
   let delta = 0;
@@ -122,7 +113,7 @@ function calculateScrollDeltaForTab(tab) {
   else if (containerRect.top > tabRect.top + offset) { // should scroll up
     delta = tabRect.top - containerRect.top + offset;
   }
-  log('calculateScrollDeltaForTab ', dumpTab(tab), {
+  log('calculateScrollDeltaForTab ', tab.id, {
     delta, offset,
     tabTop:          tabRect.top,
     tabBottom:       tabRect.bottom,
@@ -132,10 +123,11 @@ function calculateScrollDeltaForTab(tab) {
 }
 
 export function isTabInViewport(tab) {
-  if (!Tabs.ensureLivingTab(tab))
+  tab = Tab.get(tab && tab.id);
+  if (!TabsStore.ensureLivingTab(tab))
     return false;
 
-  if (Tabs.isPinned(tab))
+  if (tab.pinned)
     return true;
 
   return calculateScrollDeltaForTab(tab) == 0;
@@ -169,12 +161,12 @@ async function smoothScrollTo(params = {}) {
   const duration  = params.duration || configs.smoothScrollDuration;
   const startTime = Date.now();
 
-  return new Promise((resolve, aReject) => {
+  return new Promise((resolve, reject) => {
     const radian = 90 * Math.PI / 180;
     const scrollStep = () => {
       if (smoothScrollTo.stopped) {
         smoothScrollTo.currentOffset = 0;
-        aReject();
+        reject();
         return;
       }
       const nowTime = Date.now();
@@ -221,24 +213,25 @@ export function scrollToNewTab(tab, options = {}) {
     return;
 
   if (configs.scrollToNewTabMode == Constants.kSCROLL_TO_NEW_TAB_IF_POSSIBLE) {
-    const current = Tabs.getActiveTab();
+    const activeTab = Tab.getActiveTab(TabsStore.getWindow());
     scrollToTab(tab, Object.assign({}, options, {
-      anchor:            isTabInViewport(current) && current,
+      anchor:            !activeTab.pinned && isTabInViewport(activeTab) && activeTab,
       notifyOnOutOfView: true
     }));
   }
 }
 
 function canScrollToTab(tab) {
-  return (Tabs.ensureLivingTab(tab) &&
-          !Tabs.isHidden(tab));
+  tab = Tab.get(tab && tab.id);
+  return (TabsStore.ensureLivingTab(tab) &&
+          !tab.hidden);
 }
 
 export async function scrollToTab(tab, options = {}) {
   scrollToTab.lastTargetId = null;
 
-  log('scrollToTab to ', dumpTab(tab), dumpTab(options.anchor), options,
-      { stack: new Error().stack });
+  log('scrollToTab to ', tab && tab.id, options.anchor && options.anchor.id, options,
+      { stack: configs.debug && new Error().stack });
   cancelRunningScroll();
   if (!canScrollToTab(tab)) {
     log('=> unscrollable');
@@ -255,8 +248,8 @@ export async function scrollToTab(tab, options = {}) {
   cancelNotifyOutOfViewTab();
 
   const anchorTab = options.anchor;
-  const hasAnchor = Tabs.ensureLivingTab(anchorTab) && anchorTab != tab;
-  const openedFromPinnedTab = hasAnchor && Tabs.isPinned(anchorTab);
+  const hasAnchor = TabsStore.ensureLivingTab(anchorTab) && anchorTab != tab;
+  const openedFromPinnedTab = hasAnchor && anchorTab.pinned;
 
   if (isTabInViewport(tab) &&
       (!hasAnchor ||
@@ -272,9 +265,9 @@ export async function scrollToTab(tab, options = {}) {
   cancelNotifyOutOfViewTab();
   scrollToTab.lastTargetId = tab.id;
 
-  if (hasAnchor) {
-    const targetTabRect = tab.getBoundingClientRect();
-    const anchorTabRect = anchorTab.getBoundingClientRect();
+  if (hasAnchor && !anchorTab.pinned) {
+    const targetTabRect = tab.$TST.element.getBoundingClientRect();
+    const anchorTabRect = anchorTab.$TST.element.getBoundingClientRect();
     const containerRect = mTabBar.getBoundingClientRect();
     const offset        = getOffsetForAnimatingTab(tab);
     let delta = calculateScrollDeltaForTab(tab);
@@ -309,7 +302,7 @@ export async function scrollToTab(tab, options = {}) {
   }
   else {
     await scrollTo(Object.assign({}, options, {
-      tab: tab
+      tab
     }));
   }
   // A tab can be moved after the tabbar is scrolled to the tab.
@@ -319,9 +312,14 @@ export async function scrollToTab(tab, options = {}) {
   await wait(configs.autoGroupNewTabsTimeout);
   if (scrollToTab.stopped)
     return;
-  const retryOptions = { retryCount: options.retryCount || 0 };
+  const retryOptions = {
+    retryCount: options.retryCount || 0,
+    anchor:     options.anchor
+  };
   if (scrollToTab.lastTargetId == tab.id &&
       !isTabInViewport(tab) &&
+      (!options.anchor ||
+       !isTabInViewport(options.anchor)) &&
       retryOptions.retryCount < 3) {
     retryOptions.retryCount++;
     return scrollToTab(tab, retryOptions);
@@ -332,17 +330,21 @@ export async function scrollToTab(tab, options = {}) {
 scrollToTab.lastTargetId = null;
 
 function getOffsetForAnimatingTab(tab) {
-  const allExpanding        = document.querySelectorAll(`${Tabs.kSELECTOR_NORMAL_TAB}:not(.${Constants.kTAB_STATE_COLLAPSED}).${Constants.kTAB_STATE_EXPANDING}`);
-  const followingExpanding  = document.querySelectorAll(`#${tab.id} ~ ${Tabs.kSELECTOR_NORMAL_TAB}:not(.${Constants.kTAB_STATE_COLLAPSED}).${Constants.kTAB_STATE_EXPANDING}`);
-  const allCollapsing       = document.querySelectorAll(`${Tabs.kSELECTOR_NORMAL_TAB}.${Constants.kTAB_STATE_COLLAPSED}.${Constants.kTAB_STATE_COLLAPSING}`);
-  const followingCollapsing = document.querySelectorAll(`#${tab.id} ~ ${Tabs.kSELECTOR_NORMAL_TAB}.${Constants.kTAB_STATE_COLLAPSED}.${Constants.kTAB_STATE_COLLAPSING}`);
-  const numExpandingTabs = (allExpanding.length - followingExpanding.length) - (allCollapsing.length - followingCollapsing.length);
+  const expanding = Tab.getExpandingTabs(tab.windowId, {
+    toId:   tab.id,
+    normal: true
+  });
+  const collapsing = Tab.getCollapsingTabs(tab.windowId, {
+    toId:   tab.id,
+    normal: true
+  });
+  const numExpandingTabs = expanding.length - collapsing.length;
   return numExpandingTabs * Size.getTabHeight();
 }
 
 /*
 function scrollToTabSubtree(tab) {
-  return scrollToTab(Tabs.getLastDescendantTab(tab), {
+  return scrollToTab(tab.$TST.lastDescendant, {
     anchor:            tab,
     notifyOnOutOfView: true
   });
@@ -360,21 +362,29 @@ export function autoScrollOnMouseEvent(event) {
   if (!mTabBar.classList.contains(Constants.kTABBAR_STATE_OVERFLOW))
     return;
 
-  const tabbarRect = mTabBar.getBoundingClientRect();
-  const scrollPixels = Math.round(Size.getTabHeight() * 0.5);
-  if (event.clientY < tabbarRect.top + autoScrollOnMouseEvent.areaSize) {
-    if (mTabBar.scrollTop > 0)
-      mTabBar.scrollTop -= scrollPixels;
-  }
-  else if (event.clientY > tabbarRect.bottom - autoScrollOnMouseEvent.areaSize) {
-    if (mTabBar.scrollTop < mTabBar.scrollTopMax)
-      mTabBar.scrollTop += scrollPixels;
-  }
+  if (autoScrollOnMouseEvent.timer)
+    clearTimeout(autoScrollOnMouseEvent.timer);
+
+  autoScrollOnMouseEvent.timer = setTimeout(() => {
+    autoScrollOnMouseEvent.timer = null;
+    const tabbarRect = mTabBar.getBoundingClientRect();
+    const scrollPixels = Math.round(Size.getTabHeight() * 0.5);
+    if (event.clientY < tabbarRect.top + autoScrollOnMouseEvent.areaSize) {
+      if (mTabBar.scrollTop > 0)
+        mTabBar.scrollTop -= scrollPixels;
+    }
+    else if (event.clientY > tabbarRect.bottom - autoScrollOnMouseEvent.areaSize) {
+      if (mTabBar.scrollTop < mTabBar.scrollTopMax)
+        mTabBar.scrollTop += scrollPixels;
+    }
+  }, 0);
 }
 autoScrollOnMouseEvent.areaSize = 20;
+autoScrollOnMouseEvent.timer = null;
 
 
 async function notifyOutOfViewTab(tab) {
+  tab = Tab.get(tab && tab.id);
   if (RestoringTabCount.hasMultipleRestoringTabs()) {
     log('notifyOutOfViewTab: skip until completely restored');
     wait(100).then(() => notifyOutOfViewTab(tab));
@@ -409,14 +419,15 @@ async function onWheel(event) {
   event.stopImmediatePropagation();
   event.preventDefault();
 
+  const tab = EventUtils.getTabFromEvent(event);
   TSTAPI.notifyScrolled({
-    tab:             EventUtils.getTabFromEvent(event),
+    tab:             tab,
     scrollContainer: mTabBar,
     event:           event
   });
 }
 
-function onScroll(_aEvent) {
+function onScroll(_event) {
   reserveToSaveScrollPosition();
 }
 
@@ -426,63 +437,67 @@ function reserveToSaveScrollPosition() {
   reserveToSaveScrollPosition.reserved = setTimeout(() => {
     delete reserveToSaveScrollPosition.reserved;
     browser.sessions.setWindowValue(
-      Tabs.getWindow(),
+      TabsStore.getWindow(),
       Constants.kWINDOW_STATE_SCROLL_POSITION,
       mTabBar.scrollTop
-    );
+    ).catch(ApiTabs.createErrorSuppressor());
   }, 150);
 }
 
-
-Tabs.onCreated.addListener((tab, _info) => {
-  if (configs.animation) {
-    wait(10).then(() => { // wait until the tab is moved by TST itself
-      const parent = Tabs.getParentTab(tab);
-      if (parent && Tabs.isSubtreeCollapsed(parent)) // possibly collapsed by other trigger intentionally
-        return;
-      const active = Tabs.isActive(tab);
-      Tree.collapseExpandTab(tab, { // this is called to scroll to the tab by the "last" parameter
-        collapsed: false,
-        anchor:    Tabs.getActiveTab(),
-        last:      true
-      });
-      if (!active)
-        notifyOutOfViewTab(tab);
-    });
-  }
-  else {
-    if (Tabs.isActive(tab))
-      scrollToNewTab(tab);
-    else
-      notifyOutOfViewTab(tab);
-  }
-});
-
-Tabs.onActivated.addListener((tab, _info) => { scrollToTab(tab); });
-
-Tabs.onUnpinned.addListener(tab => { scrollToTab(tab); });
-
-
-function onMessage(message, _sender, _respond) {
-  if (!message ||
-      typeof message.type != 'string')
+function reserveToScrollToTab(tab, options = {}) {
+  if (!tab)
     return;
+  if (reserveToScrollToTab.reserved)
+    clearTimeout(reserveToScrollToTab.reserved);
+  reserveToScrollToTab.reservedTabId = tab.id;
+  reserveToScrollToTab.reservedOptions = options;
+  reserveToScrollToTab.reserved = setTimeout(() => {
+    const options = reserveToScrollToTab.reservedOptions;
+    delete reserveToScrollToTab.reservedTabId;
+    delete reserveToScrollToTab.reservedOptions;
+    delete reserveToScrollToTab.reserved;
+    scrollToTab(tab, options);
+  }, 100);
+}
 
+function reserveToScrollToNewTab(tab) {
+  if (!tab)
+    return;
+  if (reserveToScrollToNewTab.reserved)
+    clearTimeout(reserveToScrollToNewTab.reserved);
+  reserveToScrollToNewTab.reservedTabId = tab.id;
+  reserveToScrollToNewTab.reserved = setTimeout(() => {
+    delete reserveToScrollToNewTab.reservedTabId;
+    delete reserveToScrollToNewTab.reserved;
+    scrollToNewTab(tab);
+  }, 100);
+}
+
+
+function reReserveScrollingForTab(tab) {
+  if (!tab)
+    return;
+  if (reserveToScrollToTab.reservedTabId == tab.id)
+    reserveToScrollToTab(tab);
+  if (reserveToScrollToNewTab.reservedTabId == tab.id)
+    reserveToScrollToNewTab(tab);
+}
+
+
+async function onBackgroundMessage(message) {
   switch (message.type) {
-    case Constants.kCOMMAND_TAB_ATTACHED_COMPLETELY:
-      return (async () => {
-        await Tabs.waitUntilTabsAreCreated([
-          message.tab,
-          message.parent
-        ]);
-        const tab = Tabs.getTabById(message.tab);
-        if (tab && Tabs.isActive(Tabs.getTabById(message.parent)))
-          scrollToNewTab(tab);
-      })();
+    case Constants.kCOMMAND_NOTIFY_TAB_ATTACHED_COMPLETELY: {
+      await Tab.waitUntilTracked([
+        message.tabId,
+        message.parentId
+      ], { element: true });
+      const tab = Tab.get(message.tabId);
+      const parent = Tab.get(message.parentId);
+      if (tab && parent && parent.active)
+        reserveToScrollToNewTab(tab);
+    }; break;
 
     case Constants.kCOMMAND_SCROLL_TABBAR:
-      if (message.windowId != Tabs.getWindow())
-        break;
       switch (String(message.by).toLowerCase()) {
         case 'lineup':
           smoothScrollBy(-Size.getTabHeight() * configs.scrollLines);
@@ -513,6 +528,65 @@ function onMessage(message, _sender, _respond) {
           break;
       }
       break;
+
+    case Constants.kCOMMAND_NOTIFY_TAB_CREATED: {
+      await Tab.waitUntilTracked(message.tabId, { element: true });
+      if (message.maybeMoved)
+        await SidebarTabs.waitUntilNewTabIsMoved(message.tabId);
+      const tab = Tab.get(message.tabId);
+      if (configs.animation) {
+        wait(10).then(() => { // wait until the tab is moved by TST itself
+          const parent = tab.$TST.parent;
+          if (parent && parent.$TST.subtreeCollapsed) // possibly collapsed by other trigger intentionally
+            return;
+          const active = tab.active;
+          CollapseExpand.setCollapsed(tab, { // this is called to scroll to the tab by the "last" parameter
+            collapsed: false,
+            anchor:    Tab.getActiveTab(tab.windowId),
+            last:      true
+          });
+          if (!active)
+            notifyOutOfViewTab(tab);
+        });
+      }
+      else {
+        if (tab.active ||
+            Tab.getActiveTab(tab.windowId).pinned)
+          reserveToScrollToNewTab(tab);
+        else
+          notifyOutOfViewTab(tab);
+      }
+    }; break;
+
+    case Constants.kCOMMAND_NOTIFY_TAB_ACTIVATED:
+    case Constants.kCOMMAND_NOTIFY_TAB_UNPINNED:
+      await Tab.waitUntilTracked(message.tabId, { element: true });
+      reserveToScrollToTab(Tab.get(message.tabId));
+      break;
+
+    case Constants.kCOMMAND_BROADCAST_TAB_STATE: {
+      if (!message.tabIds.length ||
+          message.tabIds.length > 1 ||
+          !message.add ||
+          !message.add.includes(Constants.kTAB_STATE_BUNDLED_ACTIVE))
+        break;
+      await Tab.waitUntilTracked(message.tabIds, { element: true });
+      const tab = Tab.get(message.tabIds[0]);
+      if (!tab ||
+          tab.active)
+        break;
+      const activeTab = Tab.getActiveTab(tab.windowId);
+      reserveToScrollToTab(tab, {
+        anchor:            !activeTab.pinned && isTabInViewport(activeTab) && activeTab,
+        notifyOnOutOfView: true
+      });
+    }; break;
+
+    case Constants.kCOMMAND_NOTIFY_TAB_MOVED:
+    case Constants.kCOMMAND_NOTIFY_TAB_INTERNALLY_MOVED:
+      await Tab.waitUntilTracked(message.tabId, { element: true });
+      reReserveScrollingForTab(Tab.get(message.tabId));
+      break;
   }
 }
 
@@ -521,10 +595,10 @@ function onMessageExternal(message, _aSender) {
     case TSTAPI.kSCROLL:
       return (async () => {
         const params = {};
-        const currentWindow = Tabs.getWindow();
+        const currentWindow = TabsStore.getWindow();
         if ('tab' in message) {
-          await Tabs.waitUntilTabsAreCreated(message.tab);
-          params.tab = Tabs.getTabById(message.tab);
+          await Tab.waitUntilTracked(message.tab, { element: true });
+          params.tab = Tab.get(message.tab);
           if (!params.tab || params.tab.windowId != currentWindow)
             return;
         }
@@ -542,3 +616,25 @@ function onMessageExternal(message, _aSender) {
       })();
   }
 }
+
+CollapseExpand.onUpdating.addListener((tab, options) => {
+  if (!configs.scrollToExpandedTree)
+    return;
+  if (options.last)
+    scrollToTab(tab, {
+      anchor:            isTabInViewport(options.anchor) && options.anchor,
+      notifyOnOutOfView: true
+    });
+});
+
+CollapseExpand.onUpdated.addListener((tab, options) => {
+  if (!configs.scrollToExpandedTree)
+    return;
+  if (options.last)
+    scrollToTab(tab, {
+      anchor:            isTabInViewport(options.anchor) && options.anchor,
+      notifyOnOutOfView: true
+    });
+  else if (tab.active && !options.collapsed)
+    scrollToTab(tab);
+});
