@@ -9,12 +9,18 @@ import {
   log as internalLogger,
   configs
 } from '/common/common.js';
+import * as ApiTabs from '/common/api-tabs.js';
 import * as Constants from '/common/constants.js';
-import * as Tabs from '/common/tabs.js';
-import * as Commands from '/common/commands.js';
+import * as TabsStore from '/common/tabs-store.js';
+import * as TabsInternalOperation from '/common/tabs-internal-operation.js';
+import * as TreeBehavior from '/common/tree-behavior.js';
 import * as TSTAPI from '/common/tst-api.js';
 import * as ContextualIdentities from '/common/contextual-identities.js';
 
+import Tab from '/common/Tab.js';
+
+import * as Commands from './commands.js';
+
 import EventListenerManager from '/extlib/EventListenerManager.js';
 
 function log(...args) {
@@ -30,6 +36,12 @@ const mItemsById = {
     title:              browser.i18n.getMessage('tabContextMenu_reload_label'),
     titleMultiselected: browser.i18n.getMessage('tabContextMenu_reload_label_multiselected')
   },
+  'context_topLevel_reloadTree': {
+    title:              browser.i18n.getMessage('context_reloadTree_label')
+  },
+  'context_topLevel_reloadDescendants': {
+    title:              browser.i18n.getMessage('context_reloadDescendants_label')
+  },
   'context_toggleMuteTab-mute': {
     title:              browser.i18n.getMessage('tabContextMenu_mute_label'),
     titleMultiselected: browser.i18n.getMessage('tabContextMenu_mute_label_multiselected')
@@ -60,6 +72,9 @@ const mItemsById = {
     title:              browser.i18n.getMessage('tabContextMenu_bookmark_label'),
     titleMultiselected: browser.i18n.getMessage('tabContextMenu_bookmark_label_multiselected')
   },
+  'context_topLevel_bookmarkTree': {
+    title: browser.i18n.getMessage('context_bookmarkTree_label')
+  },
   'context_reopenInContainer': {
     title: browser.i18n.getMessage('tabContextMenu_reopenInContainer_label')
   },
@@ -91,20 +106,35 @@ const mItemsById = {
   'context_separator:afterReloadAll': {
     type: 'separator'
   },
+  'context_topLevel_collapseTree': {
+    title: browser.i18n.getMessage('context_collapseTree_label')
+  },
+  'context_topLevel_collapseAll': {
+    title: browser.i18n.getMessage('context_collapseAll_label')
+  },
+  'context_topLevel_expandTree': {
+    title: browser.i18n.getMessage('context_expandTree_label')
+  },
+  'context_topLevel_expandAll': {
+    title: browser.i18n.getMessage('context_expandAll_label')
+  },
+  'context_separator:afterCollapseExpand': {
+    type: 'separator'
+  },
   'context_closeTabsToTheEnd': {
-    title:    browser.i18n.getMessage('tabContextMenu_closeTabsToBottom_label')
+    title: browser.i18n.getMessage('tabContextMenu_closeTabsToBottom_label')
   },
   'context_closeOtherTabs': {
-    title:    browser.i18n.getMessage('tabContextMenu_closeOther_label')
+    title: browser.i18n.getMessage('tabContextMenu_closeOther_label')
   },
-  'context_closeTabOptions_closeTree': {
-    title:    browser.i18n.getMessage('context_closeTree_label')
+  'context_topLevel_closeTree': {
+    title: browser.i18n.getMessage('context_closeTree_label')
   },
-  'context_closeTabOptions_closeDescendants': {
-    title:    browser.i18n.getMessage('context_closeDescendants_label')
+  'context_topLevel_closeDescendants': {
+    title: browser.i18n.getMessage('context_closeDescendants_label')
   },
-  'context_closeTabOptions_closeOthers': {
-    title:    browser.i18n.getMessage('context_closeOthers_label')
+  'context_topLevel_closeOthers': {
+    title: browser.i18n.getMessage('context_closeOthers_label')
   },
   'context_undoCloseTab': {
     title: browser.i18n.getMessage('tabContextMenu_undoClose_label')
@@ -113,6 +143,12 @@ const mItemsById = {
     title:              browser.i18n.getMessage('tabContextMenu_close_label'),
     titleMultiselected: browser.i18n.getMessage('tabContextMenu_close_label_multiselected')
   },
+  'context_separator:afterClose': {
+    type: 'separator'
+  },
+  'context_topLevel_groupTabs': {
+    title: browser.i18n.getMessage('context_groupTabs_label')
+  },
 
   'noContextTab:context_reloadTab': {
     title: browser.i18n.getMessage('tabContextMenu_reload_label_multiselected')
@@ -159,14 +195,14 @@ export async function init() {
 
   browser.runtime.getBrowserInfo().then(browserInfo => {
     mNativeMultiselectionAvailable = parseInt(browserInfo.version.split('.')[0]) >= 63;
-  });
+  }).catch(ApiTabs.createErrorSuppressor());
 
   const itemIds = Object.keys(mItemsById);
   for (const id of itemIds) {
     const item = mItemsById[id];
     item.id          = id;
     item.lastTitle   = item.title;
-    item.lastVisible = true;
+    item.lastVisible = false;
     item.lastEnabled = true;
     if (item.type == 'separator') {
       let beforeSeparator = true;
@@ -207,6 +243,7 @@ export async function init() {
       type:     item.type || 'normal',
       contexts: ['tab'],
       viewTypes: ['sidebar'],
+      visible:  false, // hide all by default
       documentUrlPatterns: SIDEBAR_URL_PATTERN
     };
     if (item.parentId)
@@ -238,7 +275,7 @@ function updateContextualIdentities() {
     if (id in mItemsById)
       delete mItemsById[id];
     if (mNativeContextMenuAvailable)
-      browser.menus.remove(id);
+      browser.menus.remove(id).catch(ApiTabs.createErrorSuppressor());
     onExternalMessage({
       type: TSTAPI.kCONTEXT_MENU_REMOVE,
       params: id
@@ -325,7 +362,7 @@ function updateItem(id, state = {}) {
   item.lastVisible = updateInfo.visible;
   item.lastEnabled = updateInfo.enabled;
   if (mNativeContextMenuAvailable)
-    browser.menus.update(id, updateInfo);
+    browser.menus.update(id, updateInfo).catch(ApiTabs.createErrorSuppressor());
   onExternalMessage({
     type: TSTAPI.kCONTEXT_MENU_UPDATE,
     params: [id, updateInfo]
@@ -352,18 +389,16 @@ function hasVisiblePrecedingItem(separator) {
   );
 }
 
-async function onShown(info, contextApiTab) {
-  const tab                   = Tabs.getTabById(contextApiTab);
-  const windowId              = contextApiTab ? contextApiTab.windowId : (await browser.windows.getLastFocused({})).id;
-  const container             = tab ? tab.parentNode : Tabs.getTabsContainer(windowId);
-  const previousTab           = Tabs.getPreviousTab(tab);
-  const previousSiblingTab    = Tabs.getPreviousSiblingTab(tab);
-  const nextTab               = Tabs.getNextTab(tab);
-  const nextSiblingTab        = Tabs.getNextSiblingTab(tab);
-  const hasMultipleTabs       = Tabs.getTabs(tab || container).length > 1;
-  const normalTabsCount       = Tabs.getNormalTabs(tab || container).length;
-  const hasMultipleNormalTabs = normalTabsCount > 1;
-  const multiselected         = Tabs.isMultiselected(tab);
+async function onShown(info, contextTab) {
+  contextTab = contextTab && Tab.get(contextTab.id);
+  const windowId              = contextTab ? contextTab.windowId : (await browser.windows.getLastFocused({}).catch(ApiTabs.createErrorHandler())).id;
+  const previousTab           = contextTab && contextTab.$TST.previousTab;
+  const previousSiblingTab    = contextTab && contextTab.$TST.previousSiblingTab;
+  const nextTab               = contextTab && contextTab.$TST.nextTab;
+  const nextSiblingTab        = contextTab && contextTab.$TST.nextSiblingTab;
+  const hasMultipleTabs       = Tab.hasMultipleTabs(windowId);
+  const hasMultipleNormalTabs = Tab.hasMultipleTabs(windowId, { normal: true });
+  const multiselected         = contextTab && contextTab.$TST.multiselected;
 
   let modifiedItemsCount = 0;
 
@@ -375,38 +410,48 @@ async function onShown(info, contextApiTab) {
   const emulate = configs.emulateDefaultContextMenu;
 
   updateItem('context_reloadTab', {
-    visible: emulate && contextApiTab,
+    visible: emulate && contextTab,
     multiselected
   }) && modifiedItemsCount++;
+  updateItem('context_topLevel_reloadTree', {
+    visible: emulate && !multiselected && contextTab && configs.context_topLevel_reloadTree
+  }) && modifiedItemsCount++;
+  updateItem('context_topLevel_reloadDescendants', {
+    visible: emulate && !multiselected && contextTab && configs.context_topLevel_reloadDescendants,
+    enabled: contextTab && contextTab.$TST.hasChild
+  }) && modifiedItemsCount++;
   updateItem('context_toggleMuteTab-mute', {
-    visible: emulate && contextApiTab && (!contextApiTab.mutedInfo || !contextApiTab.mutedInfo.muted),
+    visible: emulate && contextTab && (!contextTab.mutedInfo || !contextTab.mutedInfo.muted),
     multiselected
   }) && modifiedItemsCount++;
   updateItem('context_toggleMuteTab-unmute', {
-    visible: emulate && contextApiTab && contextApiTab.mutedInfo && contextApiTab.mutedInfo.muted,
+    visible: emulate && contextTab && contextTab.mutedInfo && contextTab.mutedInfo.muted,
     multiselected
   }) && modifiedItemsCount++;
   updateItem('context_pinTab', {
-    visible: emulate && contextApiTab && !contextApiTab.pinned,
+    visible: emulate && contextTab && !contextTab.pinned,
     multiselected
   }) && modifiedItemsCount++;
   updateItem('context_unpinTab', {
-    visible: emulate && contextApiTab && contextApiTab.pinned,
+    visible: emulate && contextTab && contextTab.pinned,
     multiselected
   }) && modifiedItemsCount++;
   updateItem('context_duplicateTab', {
-    visible: emulate && contextApiTab,
+    visible: emulate && contextTab,
     multiselected
   }) && modifiedItemsCount++;
 
   updateItem('context_selectAllTabs', {
-    visible: emulate && contextApiTab,
-    enabled: contextApiTab && Tabs.getSelectedTabs(tab).length != Tabs.getVisibleTabs(tab).length,
+    visible: mNativeContextMenuAvailable && emulate && contextTab,
+    enabled: contextTab && Tab.getSelectedTabs(windowId).length != Tab.getVisibleTabs(windowId).length,
     multiselected
   }) && modifiedItemsCount++;
   updateItem('context_bookmarkTab', {
-    visible: emulate && contextApiTab,
-    multiselected: multiselected || !contextApiTab
+    visible: emulate && contextTab,
+    multiselected: multiselected || !contextTab
+  }) && modifiedItemsCount++;
+  updateItem('context_topLevel_bookmarkTree', {
+    visible: emulate && !multiselected && contextTab && configs.context_topLevel_bookmarkTree
   }) && modifiedItemsCount++;
 
   let showContextualIdentities = false;
@@ -416,33 +461,33 @@ async function onShown(info, contextApiTab) {
     if (!emulate)
       visible = false;
     else if (id == 'context_reopenInContainer_separator')
-      visible = contextApiTab && contextApiTab.cookieStoreId != 'firefox-default';
+      visible = contextTab && contextTab.cookieStoreId != 'firefox-default';
     else
-      visible = contextApiTab && id != `context_reopenInContainer:${contextApiTab.cookieStoreId}`;
+      visible = contextTab && id != `context_reopenInContainer:${contextTab.cookieStoreId}`;
     updateItem(id, { visible }) && modifiedItemsCount++;
     if (visible)
       showContextualIdentities = true;
   }
   updateItem('context_reopenInContainer', {
-    visible: emulate && contextApiTab && showContextualIdentities,
+    visible: emulate && contextTab && showContextualIdentities,
     multiselected
   }) && modifiedItemsCount++;
 
   updateItem('context_moveTab', {
-    visible: emulate && contextApiTab,
-    enabled: contextApiTab && hasMultipleTabs,
+    visible: emulate && contextTab,
+    enabled: contextTab && hasMultipleTabs,
     multiselected
   }) && modifiedItemsCount++;
   updateItem('context_moveTabToStart', {
-    enabled: emulate && contextApiTab && hasMultipleTabs && (previousSiblingTab || previousTab) && (Tabs.isPinned(previousSiblingTab || previousTab) == contextApiTab.pinned),
+    enabled: emulate && contextTab && hasMultipleTabs && (previousSiblingTab || previousTab) && ((previousSiblingTab || previousTab).pinned == contextTab.pinned),
     multiselected
   }) && modifiedItemsCount++;
   updateItem('context_moveTabToEnd', {
-    enabled: emulate && contextApiTab && hasMultipleTabs && (nextSiblingTab || nextTab) && (Tabs.isPinned(nextSiblingTab || nextTab) == contextApiTab.pinned),
+    enabled: emulate && contextTab && hasMultipleTabs && (nextSiblingTab || nextTab) && ((nextSiblingTab || nextTab).pinned == contextTab.pinned),
     multiselected
   }) && modifiedItemsCount++;
   updateItem('context_openTabInWindow', {
-    enabled: emulate && contextApiTab && hasMultipleTabs,
+    enabled: emulate && contextTab && hasMultipleTabs,
     multiselected
   }) && modifiedItemsCount++;
 
@@ -454,82 +499,96 @@ async function onShown(info, contextApiTab) {
     visible: emulate && !mNativeMultiselectionAvailable
   }) && modifiedItemsCount++;
 
+  updateItem('context_topLevel_collapseTree', {
+    visible: emulate && !multiselected && contextTab && configs.context_topLevel_collapseTree,
+    enabled: contextTab && contextTab.$TST.hasChild
+  }) && modifiedItemsCount++;
+  updateItem('context_topLevel_collapseAll', {
+    visible: emulate && !multiselected && contextTab && configs.context_topLevel_collapseAll
+  }) && modifiedItemsCount++;
+  updateItem('context_topLevel_expandTree', {
+    visible: emulate && !multiselected && contextTab && configs.context_topLevel_expandTree,
+    enabled: contextTab && contextTab.$TST.hasChild
+  }) && modifiedItemsCount++;
+  updateItem('context_topLevel_expandAll', {
+    visible: emulate && !multiselected && contextTab && configs.context_topLevel_expandAll
+  }) && modifiedItemsCount++;
+
   updateItem('context_closeTabsToTheEnd', {
-    visible: emulate && contextApiTab,
+    visible: emulate && contextTab,
     enabled: hasMultipleNormalTabs && nextTab,
     multiselected
   }) && modifiedItemsCount++;
   updateItem('context_closeOtherTabs', {
-    visible: emulate && contextApiTab,
+    visible: emulate && contextTab,
     enabled: hasMultipleNormalTabs,
     multiselected
   }) && modifiedItemsCount++;
 
-  updateItem('context_closeTabOptions_closeTree', {
-    visible: emulate && contextApiTab && configs.context_closeTabOptions_closeTree,
-    enabled: !multiselected
+  updateItem('context_topLevel_closeTree', {
+    visible: emulate && !multiselected && contextTab && configs.context_topLevel_closeTree
   }) && modifiedItemsCount++;
-  updateItem('context_closeTabOptions_closeDescendants', {
-    visible: emulate && contextApiTab && configs.context_closeTabOptions_closeDescendants,
-    enabled: !multiselected && Tabs.hasChildTabs(tab)
+  updateItem('context_topLevel_closeDescendants', {
+    visible: emulate && !multiselected && contextTab && configs.context_topLevel_closeDescendants,
+    enabled: contextTab && contextTab.$TST.hasChild
   }) && modifiedItemsCount++;
-  updateItem('context_closeTabOptions_closeOthers', {
-    visible: emulate && contextApiTab && configs.context_closeTabOptions_closeOthers,
-    enabled: !multiselected
+  updateItem('context_topLevel_closeOthers', {
+    visible: emulate && !multiselected && contextTab && configs.context_topLevel_closeOthers
   }) && modifiedItemsCount++;
 
   updateItem('context_undoCloseTab', {
-    visible: emulate && contextApiTab,
+    visible: emulate && contextTab,
     multiselected
   }) && modifiedItemsCount++;
   updateItem('context_closeTab', {
-    visible: emulate && contextApiTab,
+    visible: emulate && contextTab,
     multiselected
   }) && modifiedItemsCount++;
 
   updateItem('noContextTab:context_reloadTab', {
-    visible: emulate && !contextApiTab
+    visible: emulate && !contextTab
   }) && modifiedItemsCount++;
   updateItem('noContextTab:context_bookmarkSelected', {
-    visible: emulate && !contextApiTab && mNativeMultiselectionAvailable
+    visible: emulate && !contextTab && mNativeMultiselectionAvailable
   }) && modifiedItemsCount++;
   updateItem('noContextTab:context_selectAllTabs', {
-    visible: emulate && !contextApiTab,
-    enabled: !contextApiTab && Tabs.getSelectedTabs(tab).length != Tabs.getVisibleTabs(tab).length
+    visible: emulate && !contextTab,
+    enabled: !contextTab && Tab.getSelectedTabs(windowId).length != Tab.getVisibleTabs(windowId).length
   }) && modifiedItemsCount++;
   updateItem('noContextTab:context_undoCloseTab', {
-    visible: emulate && !contextApiTab
+    visible: emulate && !contextTab
+  }) && modifiedItemsCount++;
+
+  updateItem('context_topLevel_groupTabs', {
+    visible: emulate && multiselected && contextTab && configs.context_topLevel_groupTabs
   }) && modifiedItemsCount++;
 
   updateSeparator('context_separator:afterDuplicate') && modifiedItemsCount++;
   updateSeparator('context_separator:afterSendTab') && modifiedItemsCount++;
   updateSeparator('context_separator:afterReloadAll') && modifiedItemsCount++;
+  updateSeparator('context_separator:afterCollapseExpand') && modifiedItemsCount++;
+  updateSeparator('context_separator:afterClose') && modifiedItemsCount++;
 
-  let flattenExtraItems;
-  if (typeof Array.prototype.flat !== 'function') // Firefox 61 or older
-    flattenExtraItems = Array.from(mExtraItems.values()).reduce((acc, val) => acc.concat(val), []);
-  else
-    flattenExtraItems = Array.from(mExtraItems.values()).flat();
+  const flattenExtraItems = Array.from(mExtraItems.values()).flat();
 
   updateSeparator('lastSeparatorBeforeExtraItems', {
-    hasVisibleFollowing: contextApiTab && flattenExtraItems.some(item => !item.parentId && item.visible !== false)
+    hasVisibleFollowing: contextTab && flattenExtraItems.some(item => !item.parentId && item.visible !== false)
   }) && modifiedItemsCount++;
 
   /* eslint-enable no-unused-expressions */
 
   if (mNativeContextMenuAvailable && modifiedItemsCount)
-    browser.menus.refresh();
+    browser.menus.refresh().catch(ApiTabs.createErrorSuppressor());
 }
 
-async function onClick(info, contextApiTab) {
-  const window            = await browser.windows.getLastFocused({ populate: true });
-  const contextWindowId   = window.id;
-  const contextTabElement = Tabs.getTabById(contextApiTab);
-  const activeTab         = window.tabs.find(tab => tab.active);
-  const activeTabElement  = Tabs.getTabById(activeTab);
+async function onClick(info, contextTab) {
+  contextTab = contextTab && Tab.get(contextTab.id);
+  const window    = await browser.windows.getLastFocused({ populate: true }).catch(ApiTabs.createErrorHandler());
+  const windowId  = contextTab && contextTab.windowId || window.id;
+  const activeTab = TabsStore.activeTabInWindow.get(windowId);
 
-  let multiselectedTabs = Tabs.getSelectedTabs(contextTabElement || activeTabElement);
-  const isMultiselected = contextTabElement ? Tabs.isMultiselected(contextTabElement) : multiselectedTabs.length > 1;
+  let multiselectedTabs = Tab.getSelectedTabs(windowId);
+  const isMultiselected = contextTab ? contextTab.$TST.multiselected : multiselectedTabs.length > 1;
   if (!isMultiselected)
     multiselectedTabs = null;
 
@@ -537,156 +596,169 @@ async function onClick(info, contextApiTab) {
     case 'context_reloadTab':
       if (multiselectedTabs) {
         for (const tab of multiselectedTabs) {
-          browser.tabs.reload(tab.apiTab.id);
+          browser.tabs.reload(tab.id)
+            .catch(ApiTabs.createErrorHandler(ApiTabs.handleMissingTabError));
         }
       }
       else {
-        const tab = contextTabElement || Tabs.getActiveTab(contextWindowId);
-        browser.tabs.reload(tab.apiTab.id);
+        const tab = contextTab || activeTab;
+        browser.tabs.reload(tab.id)
+          .catch(ApiTabs.createErrorHandler(ApiTabs.handleMissingTabError));
       }
       break;
     case 'context_toggleMuteTab-mute':
       if (multiselectedTabs) {
         for (const tab of multiselectedTabs) {
-          browser.tabs.update(tab.apiTab.id, { muted: true });
+          browser.tabs.update(tab.id, { muted: true })
+            .catch(ApiTabs.createErrorHandler(ApiTabs.handleMissingTabError));
         }
       }
       else {
-        browser.tabs.update(contextApiTab.id, { muted: true });
+        browser.tabs.update(contextTab.id, { muted: true })
+          .catch(ApiTabs.createErrorHandler(ApiTabs.handleMissingTabError));
       }
       break;
     case 'context_toggleMuteTab-unmute':
       if (multiselectedTabs) {
         for (const tab of multiselectedTabs) {
-          browser.tabs.update(tab.apiTab.id, { muted: false });
+          browser.tabs.update(tab.id, { muted: false })
+            .catch(ApiTabs.createErrorHandler(ApiTabs.handleMissingTabError));
         }
       }
       else {
-        browser.tabs.update(contextApiTab.id, { muted: false });
+        browser.tabs.update(contextTab.id, { muted: false })
+          .catch(ApiTabs.createErrorHandler(ApiTabs.handleMissingTabError));
       }
       break;
     case 'context_pinTab':
       if (multiselectedTabs) {
         for (const tab of multiselectedTabs) {
-          browser.tabs.update(tab.apiTab.id, { pinned: true });
+          browser.tabs.update(tab.id, { pinned: true })
+            .catch(ApiTabs.createErrorHandler(ApiTabs.handleMissingTabError));
         }
       }
       else {
-        browser.tabs.update(contextApiTab.id, { pinned: true });
+        browser.tabs.update(contextTab.id, { pinned: true })
+          .catch(ApiTabs.createErrorHandler(ApiTabs.handleMissingTabError));
       }
       break;
     case 'context_unpinTab':
       if (multiselectedTabs) {
         for (const tab of multiselectedTabs) {
-          browser.tabs.update(tab.apiTab.id, { pinned: false });
+          browser.tabs.update(tab.id, { pinned: false })
+            .catch(ApiTabs.createErrorHandler(ApiTabs.handleMissingTabError));
         }
       }
       else {
-        browser.tabs.update(contextApiTab.id, { pinned: false });
+        browser.tabs.update(contextTab.id, { pinned: false })
+          .catch(ApiTabs.createErrorHandler(ApiTabs.handleMissingTabError));
       }
       break;
     case 'context_duplicateTab':
-      Commands.duplicateTab(contextTabElement, {
-        destinationWindowId: contextWindowId
+      Commands.duplicateTab(contextTab, {
+        destinationWindowId: windowId
       });
       break;
     case 'context_moveTabToStart':
-      Commands.moveTabToStart(contextTabElement);
+      Commands.moveTabToStart(contextTab);
       break;
     case 'context_moveTabToEnd':
-      Commands.moveTabToEnd(contextTabElement);
+      Commands.moveTabToEnd(contextTab);
       break;
     case 'context_openTabInWindow':
-      Commands.openTabInWindow(contextTabElement);
+      Commands.openTabInWindow(contextTab);
+      break;
     case 'context_selectAllTabs': {
-      const apiTabs = await browser.tabs.query({ windowId: contextWindowId });
+      const tabs = await browser.tabs.query({ windowId }).catch(ApiTabs.createErrorHandler());
       browser.tabs.highlight({
-        windowId: contextWindowId,
+        windowId,
         populate: false,
-        tabs:     [activeTab.index].concat(apiTabs.filter(tab => !tab.active).map(tab => tab.index))
-      });
+        tabs:     [activeTab.index].concat(tabs.filter(tab => !tab.active).map(tab => tab.index))
+      }).catch(ApiTabs.createErrorSuppressor());
     }; break;
     case 'context_bookmarkTab':
-      Commands.bookmarkTab(contextTabElement);
+      Commands.bookmarkTab(contextTab);
       break;
     case 'context_bookmarkSelected':
-      Commands.bookmarkTab(contextTabElement || activeTabElement);
+      Commands.bookmarkTab(contextTab || activeTab);
       break;
     case 'context_bookmarkAllTabs':
-      Commands.bookmarkTabs(Tabs.getTabs(contextTabElement));
+      Commands.bookmarkTabs(Tab.getTabs(contextTab.windowId));
       break;
     case 'context_reloadAllTabs': {
-      const apiTabs = await browser.tabs.query({ windowId: contextWindowId }) ;
-      for (const apiTab of apiTabs) {
-        browser.tabs.reload(apiTab.id);
+      const tabs = await browser.tabs.query({ windowId }).catch(ApiTabs.createErrorHandler());
+      for (const tab of tabs) {
+        browser.tabs.reload(tab.id)
+          .catch(ApiTabs.createErrorHandler(ApiTabs.handleMissingTabError));
       }
     }; break;
     case 'context_closeTabsToTheEnd': {
-      const apiTabs = await browser.tabs.query({ windowId: contextWindowId });
+      const tabs = await browser.tabs.query({ windowId }).catch(ApiTabs.createErrorHandler());
       let after = false;
-      const closeApiTabIds = [];
+      const closeTabs = [];
       const keptTabIds = new Set(
         multiselectedTabs ?
-          multiselectedTabs.map(tab => tab.apiTab.id) :
-          [contextApiTab.id]
+          multiselectedTabs.map(tab => tab.id) :
+          [contextTab.id]
       );
-      for (const apiTab of apiTabs) {
-        if (keptTabIds.has(apiTab.id)) {
+      for (const tab of tabs) {
+        if (keptTabIds.has(tab.id)) {
           after = true;
           continue;
         }
-        if (after && !apiTab.pinned)
-          closeApiTabIds.push(apiTab.id);
+        if (after && !tab.pinned)
+          closeTabs.push(Tab.get(tab.id));
       }
       const canceled = (await browser.runtime.sendMessage({
-        type:     Constants.kCOMMAND_NOTIFY_TABS_CLOSING,
-        tabs:     closeApiTabIds,
-        windowId: contextWindowId
-      })) === false
+        type: Constants.kCOMMAND_NOTIFY_TABS_CLOSING,
+        tabs: closeTabs.map(tab => tab.$TST.sanitized),
+        windowId
+      }).catch(ApiTabs.createErrorHandler())) === false;
       if (canceled)
         break;
-      browser.tabs.remove(closeApiTabIds);
+      TabsInternalOperation.removeTabs(closeTabs);
     }; break;
     case 'context_closeOtherTabs': {
-      const apiTabs  = await browser.tabs.query({ windowId: contextWindowId });
+      const tabs  = await browser.tabs.query({ windowId }).catch(ApiTabs.createErrorHandler());
       const keptTabIds = new Set(
         multiselectedTabs ?
-          multiselectedTabs.map(tab => tab.apiTab.id) :
-          [contextApiTab.id]
+          multiselectedTabs.map(tab => tab.id) :
+          [contextTab.id]
       );
-      const closeApiTabIds = apiTabs.filter(apiTab => !apiTab.pinned && !keptTabIds.has(apiTab.id)).map(tab => tab.id);
+      const closeTabs = tabs.filter(tab => !tab.pinned && !keptTabIds.has(tab.id)).map(tab => Tab.get(tab.id));
       const canceled = (await browser.runtime.sendMessage({
-        type:     Constants.kCOMMAND_NOTIFY_TABS_CLOSING,
-        tabs:     closeApiTabIds,
-        windowId: contextWindowId
-      })) === false
+        type: Constants.kCOMMAND_NOTIFY_TABS_CLOSING,
+        tabs: closeTabs.map(tab => tab.$TST.sanitized),
+        windowId
+      }).catch(ApiTabs.createErrorHandler())) === false;
       if (canceled)
         break;
-      browser.tabs.remove(closeApiTabIds);
+      TabsInternalOperation.removeTabs(closeTabs);
     }; break;
     case 'context_undoCloseTab': {
-      const sessions = await browser.sessions.getRecentlyClosed({ maxResults: 1 });
+      const sessions = await browser.sessions.getRecentlyClosed({ maxResults: 1 }).catch(ApiTabs.createErrorHandler());
       if (sessions.length && sessions[0].tab)
-        browser.sessions.restore(sessions[0].tab.sessionId);
+        browser.sessions.restore(sessions[0].tab.sessionId).catch(ApiTabs.createErrorSuppressor());
+    }; break;
+    case 'context_closeTab': {
+      const closeTabs = (multiselectedTabs || TreeBehavior.getClosingTabsFromParent(contextTab, {
+        byInternalOperation: true
+      })).reverse(); // close down to top, to keep tree structure of Tree Style Tab
+      const canceled = (await browser.runtime.sendMessage({
+        type: Constants.kCOMMAND_NOTIFY_TABS_CLOSING,
+        tabs: closeTabs.map(tab => tab.$TST.sanitized),
+        windowId
+      }).catch(ApiTabs.createErrorHandler())) === false;
+      if (canceled)
+        return;
+      TabsInternalOperation.removeTabs(closeTabs);
     }; break;
-    case 'context_closeTab':
-      if (multiselectedTabs) {
-        // close down to top, to keep tree structure of Tree Style Tab
-        multiselectedTabs.reverse();
-        for (const tab of multiselectedTabs) {
-          browser.tabs.remove(tab.apiTab.id);
-        }
-      }
-      else {
-        browser.tabs.remove(contextApiTab.id);
-      }
-      break;
 
     default: {
       const contextualIdentityMatch = info.menuItemId.match(/^context_reopenInContainer:(.+)$/);
-      if (contextApiTab &&
+      if (contextTab &&
           contextualIdentityMatch)
-        Commands.reopenInContainer(contextTabElement, contextualIdentityMatch[1]);
+        Commands.reopenInContainer(contextTab, contextualIdentityMatch[1]);
     }; break;
   }
 }
@@ -713,7 +785,7 @@ async function notifyUpdated() {
   await browser.runtime.sendMessage({
     type:  TSTAPI.kCONTEXT_MENU_UPDATED,
     items: exportExtraItems()
-  }).catch(_error => {});
+  }).catch(ApiTabs.createErrorSuppressor());
 }
 
 let mReservedNotifyUpdate;
